"""Definition of the CatTransform."""
import transform as n
from wtrwrks.waterworks.empty import empty
import wtrwrks.tanks.tank_defs as td
import wtrwrks.read_write.tf_features as feat
import numpy as np
import logging
import tensorflow as tf


class CatTransform(n.Transform):
  """Class used to create mappings from raw categorical to vectorized, normalized data and vice versa.

  Parameters
  ----------
  name : str
    The name of the transform.
  dtype : numpy dtype
    The data type the transformed data should have. Defaults to np.float64.
  input_dtype: numpy dtype
    The datatype of the original inputted array.
  norm_mode : "mean_std" or None
    How to normalize the data. Subtracting out the mean and dividing by the standard deviation or leaving as is.
  norm_axis : 0, 1, or None
    In the case that 'norm_mode' has a non None value, what axis should be used for the normalization. None implies both 0 and 1.
  index_to_cat_val : list
    The mapping from index number to category value.

  Attributes
  ----------
  attribute_dict: dict
    The keys are the attributes of the class while the values are the default values. It's done this way rather than defined in the __init__ because this dictionary also defines what values need to be saved when written to disk, and what values need to be displayed when printed to the terminal.
  required_params: set of strs
    The parameters that must be provided to the transform at definition.
  cols : list of strs
    The column names of the array or dataframe to be transformed.
  num_examples : int
    The number of rows that have passed through the calc_global_values function.
  is_calc_run : bool
    Whether or not calc_global_values has been run for this transform.
  mean : numpy array
    The stored mean values to be used to normalize data.
  std : numpy array
    The stored standard deviation values to be used to normalize data.
  cat_val_to_index : dict
    The mapping from category value to index number.

  """

  attribute_dict = {'norm_mode': None, 'norm_axis': 0, 'name': '', 'mean': None, 'std': None, 'dtype': np.float64, 'index_to_cat_val': None, 'cat_val_to_index': None}
  for k, v in n.Transform.attribute_dict.iteritems():
    if k in attribute_dict:
      continue
    attribute_dict[k] = v

  required_params = set(['index_to_cat_val'])
  required_params.update(n.Transform.required_params)

  def __init__(self, from_file=None, save_dict=None, **kwargs):
    """Define a transform using a dictionary, file, or by setting the attribute values in kwargs.

    Parameters
    ----------
    from_file : None or str
      The file path of the Tranform that was written to disk.
    save_dict : dict or None
      The dictionary of attributes that completely define a Transform.
    name : str
      The name of the transform.
    dtype : numpy dtype
      The data type the transformed data should have. Defaults to np.float64.
    input_dtype: numpy dtype
      The datatype of the original inputted array.
    norm_mode : "mean_std" or None
      How to normalize the data. Subtracting out the mean and dividing by the standard deviation or leaving as is.
    norm_axis : 0, 1, or None
      In the case that 'norm_mode' has a non None value, what axis should be used for the normalization. None implies both 0 and 1.
    index_to_cat_val : list
      The mapping from index number to category value.
    **kwargs :
      The keyword arguments that set the values of the attributes defined in the attribute_dict.

    """
    super(CatTransform, self).__init__(from_file, save_dict, **kwargs)

    # Check index to cat value are all unique.
    if len(self.index_to_cat_val) != len(set(self.index_to_cat_val)):
      raise ValueError("All elements of index_to_cat_val must be unique")

    # Ensure a valid norm mode was passed
    valid_norm_modes = ('mean_std', None)
    if self.norm_mode not in valid_norm_modes:
      raise ValueError("{} is an invalid norm_mode. Accepted norm mods are ".format(self.norm_axis, valid_norm_modes))

    # Ensure a valid axis was passed
    valid_norm_axis = (0, 1, (0, 1), None)
    if self.norm_axis not in valid_norm_axis:
      raise ValueError("{} is an invalid norm_axis. Accepted norm axes are ".format(self.norm_mode, valid_norm_axis))

  def __len__(self):
    """Get the length of the transformed data"""
    # assert self.is_calc_run, ("Must run calc_global_values before taking the len.")
    return len(self.index_to_cat_val)

  def _calc_global_values(self, array):
    """Calculate all the values of the Transform that are dependent on all the examples of the dataset. (e.g. mean, standard deviation, unique category values, etc.) This method must be run before any actual transformation can be done.

    Parameters
    ----------
    array : np.ndarray
      Some of the data that will be transformed.

    """
    if self.input_dtype is None:
      self.input_dtype = array.dtype
    else:
      array = np.array(array, dtype=self.input_dtype)

    batch_size = float(array.shape[0])
    total_examples = self.num_examples + batch_size

    if self.norm_mode == 'mean_std':
      if not len(self.index_to_cat_val):
        raise ValueError("index_to_cat_val has no valid values.")

      # Find all the category values which are not known
      valid_cats = np.isin(array, self.index_to_cat_val)
      array = np.array(array, copy=True)

      # Replace the unknown category values with some default value.
      default_val = self.index_to_cat_val[0]
      if isinstance(default_val, float) and np.isnan(default_val):
        default_val = self.index_to_cat_val[1]
      array[~valid_cats] = default_val

      one_hots = np.zeros(list(array.shape) + [len(self.index_to_cat_val)], dtype=np.float64)

      # Convert the category values to indices, using the cat_vat_to_index.
      indices = np.vectorize(self.cat_val_to_index.get)(array)

      # Generate all the indices of the 'indices' array. i.e. one_hot_indices is
      # equal to something like (0, 0...0), (0, 0 ... 1)... (n1, n2... nk).
      # Where n1..nk are the size of each dimension of 'indices'. Then append
      # the corresponding value from 'indices' to get the full location of where
      # a 1 should be in the one_hots array.
      one_hot_indices = np.unravel_index(np.arange(indices.size, dtype=np.int32), indices.shape)
      one_hot_indices = list(one_hot_indices) + [indices.flatten()]

      # Set all the proper locations to 1. And then undo the setting of the
      # not valid categories.
      # one_hot_indices = np.array(one_hot_indices, dtype=np.int32)
      one_hots[one_hot_indices] = 1
      one_hots[~valid_cats] = 0

      # Get the mean and standard deviation of the one_hots.
      if self.mean is None:
        self.mean = np.mean(one_hots, axis=self.norm_axis)
        self.var = np.var(one_hots, axis=self.norm_axis)
      else:
        self.mean = (self.num_examples / total_examples) * self.mean + (batch_size / total_examples) * np.mean(one_hots, axis=self.norm_axis)
        self.var = (self.num_examples / total_examples) * self.mean + (batch_size / total_examples) * np.var(one_hots, axis=self.norm_axis)

    self.num_examples += batch_size

  def _finish_calc(self):
    """Finish up the calc global value process."""
    if self.norm_mode == 'mean_std':
      self.std = np.sqrt(self.var)

      # If there are any standard deviations of 0, replace them with 1's,
      # print out a warning.
      if len(self.std[self.std == 0]):
        zero_std_cat_vals = []
        for index in np.where(self.std == 0.0)[0]:
          zero_std_cat_vals.append(self.index_to_cat_val[index])

        logging.warn(self.name + " has zero-valued stds at " + str(zero_std_cat_vals) + " replacing with 1's")

        self.std[self.std == 0] = 1.0

  def _get_array_attributes(self, prefix=''):
    """Get the dictionary that contain the original shapes of the arrays before being converted into tfrecord examples.

    Parameters
    ----------
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    array_attributes : dict
      The dictionary with keys equal to those that are found in the Transform's example dicts and values are the shapes of the arrays of a single example.

    """
    att_dict = {}
    att_dict['missing_vals'] = {
      'shape': [len(self.cols)],
      'tf_type': feat.select_tf_dtype(self.input_dtype),
      'size': feat.size_from_shape([len(self.cols)]),
      'feature_func': feat.select_feature_func(self.input_dtype),
      'np_type': self.input_dtype
    }
    one_hots_shape = [len(self.cols)] + [len(self.index_to_cat_val)]
    att_dict['one_hots'] = {
      'shape': one_hots_shape,
      'tf_type': feat.select_tf_dtype(self.dtype),
      'size': feat.size_from_shape(one_hots_shape),
      'feature_func': feat.select_feature_func(self.dtype),
      'np_type': self.dtype
    }
    att_dict['indices'] = {
      'shape': [len(self.cols)],
      'tf_type': tf.int64,
      'size': feat.size_from_shape([len(self.cols)]),
      'feature_func': feat._int_feat,
      'np_type': np.int64
    }

    att_dict = self._pre(att_dict, prefix)
    return att_dict

  def _start_calc(self):
    """Start the calc global value process."""
    # Create the mapping from category values to index in the vector and
    # vice versa
    self.cat_val_to_index = {}

    for unique_num, unique in enumerate(self.index_to_cat_val):
      cat_val = self.index_to_cat_val[unique_num]
      self.cat_val_to_index[cat_val] = unique_num
    self.num_examples = 0.

  def define_waterwork(self, array=empty, return_tubes=None, prefix=''):
    """Get the waterwork that completely describes the pour and pump transformations.

    Parameters
    ----------
    array : np.ndarray or empty
      The array to be transformed.

    Returns
    -------
    Waterwork
      The waterwork with all the tanks (operations) added, and names set.

    """
    # Convert the category values to indices.
    cti, cti_slots = td.cat_to_index(
      array, self.cat_val_to_index,
      tube_plugs={'input_dtype': lambda z: self.input_dtype}
    )
    cti_slots['cats'].set_name('array')
    cti['missing_vals'].set_name('missing_vals')

    # Clone the indices so that a copy of 'indices' can be outputted as a tap.
    cloned, _ = td.clone(cti['target'])
    cloned['a'].set_name('indices')

    # Convert the indices into one-hot vectors.
    one_hots, _ = td.one_hot(
      cloned['b'], len(self.cat_val_to_index),
      tube_plugs={
        'missing_vals': lambda z: np.ones(z[self._pre('indices', prefix)].shape)*-2
      }
    )

    if self.norm_mode == 'mean_std':
      one_hots, _ = td.sub(
        one_hots['target'], self.mean,
        tube_plugs={'a_is_smaller': False, 'smaller_size_array': self.mean}
      )
      one_hots, _ = td.div(
        one_hots['target'], self.std,
        tube_plugs={'a_is_smaller': False, 'smaller_size_array': self.std, 'missing_vals': np.array([]), 'remainder': np.array([])}
      )

    one_hots['target'].set_name('one_hots')

    if return_tubes is not None:
      ww = one_hots['target'].waterwork
      r_tubes = []
      for r_tube_key in return_tubes:
        r_tubes.append(ww.maybe_get_tube(r_tube_key))
      return r_tubes

  def get_schema_dict(self, var_lim=None):
    """Create a dictionary which defines the proper fields which are needed to store the untransformed data in a (postgres) SQL database.

    Parameters
    ----------
    var_lim : int or dict
      The maximum size of strings. If an int is passed then all VARCHAR fields have the same limit. If a dict is passed then each field gets it's own limit. Defaults to 255 for all string fields.

    Returns
    -------
    schema_dict : dict
      Dictionary where the keys are the field names and the values are the SQL data types.

    """
    if var_lim is None:
      if self.input_dtype in (np.dtype('U'), np.dtype('S'), np.dtype('O')):
        var_lim = 1
        for cat_val in self.index_to_cat_val:
          if len(cat_val) > var_lim:
            var_lim = len(cat_val)
      else:
        var_lim = 255

    return super(CatTransform, self).get_schema_dict(var_lim)



import shutil
import tempfile
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.transforms.cat_transform as n
import numpy as np
import pandas as pd


class TestCatTransform(th.TestTransform):
  def setUp(self):
      self.temp_dir = tempfile.mkdtemp()
      self.array = np.array([
        ['a', 'b', 1.0],
        ['b', 'None', 2.0],
        ['c', 'b', np.nan],
        ['a', 'c', 1.0],
      ], dtype=np.object)
      self.index_to_cat_val = ['a', 'b']

  def tearDown(self):
      shutil.rmtree(self.temp_dir)

  def test_no_norm(self):
    trans = n.CatTransform(
      name='cat',
      index_to_cat_val=np.unique(self.array[:, 0: 1].astype(np.str))
    )
    trans.calc_global_values(self.array[:, 0: 1].astype(np.str))
    target = np.array([
      [[1., 0., 0.]],
      [[0., 1., 0.]],
      [[0., 0., 1.]],
      [[1., 0., 0.]]
    ])
    indices = np.array([[0], [1], [2], [0]])
    for i in xrange(2):
      self.pour_pump(
        trans,
        self.array[:, 0: 1].astype(np.str),
        {
          'cat/missing_vals': [[''], [''], [''], ['']],
          'cat/one_hots': target,
          'cat/indices': indices
        }
      )
      trans = self.write_read(trans, self.temp_dir)

  def test_two_cols(self):
    trans = n.CatTransform(
      name='cat',
      index_to_cat_val=np.unique(self.array[:, 0: 2].astype(np.str))
    )
    array_iter = [self.array[0:2, 0: 2].astype(str), self.array[2: 4, 0: 2].astype(str)]

    trans.calc_global_values(data_iter=array_iter)
    target = np.array([
      [[0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]],
      [[0.0, 0.0, 1.0, 0.0], [1.0, 0.0, 0.0, 0.0]],
      [[0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 1.0, 0.0]],
      [[0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0]]
    ])
    indices = np.array([[1, 2], [2, 0], [3, 2], [1, 3]])
    for i in xrange(2):
      self.pour_pump(
        trans,
        self.array[:, 0: 2].astype(np.str),
        {
          'cat/missing_vals': np.array([['', ''], ['', ''], ['', ''], ['', '']], dtype='|S4'),
          'cat/one_hots': target,
          'cat/indices': indices
        }
      )
      trans = self.write_read(trans, self.temp_dir)

  def test_df_iter(self):
    trans = n.CatTransform(
      name='cat',
      index_to_cat_val=np.unique(self.array[:, 0: 2].astype(np.str)),
      input_dtype=np.dtype('S')
    )
    array_iter = [self.array[0:2, 0: 2].astype(str), self.array[2: 4, 0: 2].astype(str)]
    df_iter = []
    for array in array_iter:
      df = pd.DataFrame(data=array, index=[0, 1], columns=['a', 'b'])
      df_iter.append(df)

    trans.calc_global_values(data_iter=df_iter)
    target = np.array([
      [[0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]],
      [[0.0, 0.0, 1.0, 0.0], [1.0, 0.0, 0.0, 0.0]],
      [[0.0, 0.0, 0.0, 1.0], [0.0, 0.0, 1.0, 0.0]],
      [[0.0, 1.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0]]
    ])
    indices = np.array([[1, 2], [2, 0], [3, 2], [1, 3]])
    for i in xrange(2):
      self.pour_pump(
        trans,
        self.array[:, 0: 2].astype(np.str),
        {
          'cat/missing_vals': np.array([['', ''], ['', ''], ['', ''], ['', '']], dtype='|S4'),
          'cat/one_hots': target,
          'cat/indices': indices
        }
      )
      trans = self.write_read(trans, self.temp_dir)

  def test_index_to_cat_val(self):
    trans = n.CatTransform(
      name='cat',
      index_to_cat_val=['a', 'b']
    )
    trans.calc_global_values(self.array[:, 0: 1].astype(np.str))
    target = np.array([[[1.0, 0.0]], [[0.0, 1.0]], [[0.0, 0.0]], [[1.0, 0.0]]]).astype(float)
    indices = np.array([[0], [1], [-1], [0]])
    for i in xrange(2):
      self.pour_pump(
        trans,
        self.array[:, 0: 1].astype(np.str),
        {
          'cat/missing_vals': [[''], [''], ['c'], ['']],
          'cat/one_hots': target,
          'cat/indices': indices
        }
      )
      trans = self.write_read(trans, self.temp_dir)

  def test_null(self):
    trans = n.CatTransform(
      name='cat',
      index_to_cat_val=np.unique(self.array[:, 2: 3].astype(np.float64))
    )
    trans.calc_global_values(self.array[:, 2: 3].astype(np.float64))
    target = np.array([[[1.0, 0.0, 0.0]], [[0.0, 1.0, 0.0]], [[0.0, 0.0, 1.0]], [[1.0, 0.0, 0.0]]]).astype(float)
    indices = np.array([[0], [1], [2], [0]])

    for i in xrange(2):
      self.pour_pump(
        trans,
        self.array[:, 2: 3].astype(np.float64),
        {
          'cat/missing_vals': np.array([[0.0], [0.0], [0.0], [0.0]], dtype=float),
          'cat/one_hots': target,
          'cat/indices': indices
        }
      )
      trans = self.write_read(trans, self.temp_dir)

  def test_mean_std(self):
    trans = n.CatTransform(
      name='cat',
      norm_mode='mean_std',
      index_to_cat_val=sorted(np.unique(self.array[:, 1: 2].astype(np.str)))
    )

    array_iter = [self.array[0:2, 1: 2].astype(str), self.array[2: 4, 1: 2].astype(str)]

    trans.calc_global_values(data_iter=array_iter)
    target = np.array([
          [[0., 1, 0]],
          [[1, 0, 0]],
          [[0, 1, 0]],
          [[0, 0, 1]],
        ]).astype(float)
    # print trans.mean
    target = (target - trans.mean)/trans.std
    indices = np.array([[1], [0], [1], [2]])
    for i in xrange(2):
      self.pour_pump(
        trans,
        self.array[:, 1: 2].astype(np.str),
        {
          'cat/missing_vals': np.array([[''], [''], [''], ['']], dtype='|S4'),
          'cat/one_hots': target,
          'cat/indices': indices
        }
      )
      trans = self.write_read(trans, self.temp_dir)

  # def test_read_write(self):
  #
  #   for i in xrange(3):
  #     if i in (0, 1):
  #       array = self.array[:, i: i + 1].astype(np.str)
  #     else:
  #       array = self.array[:, i: i + 1].astype(np.float)
  #     trans = n.CatTransform(
  #       name='cat',
  #       norm_mode='mean_std',
  #       index_to_cat_val=np.unique(array)
  #     )
  #     trans.calc_global_values(array)
  #     self.write_read_example(trans, array, self.temp_dir)

  def test_errors(self):
    with self.assertRaises(ValueError):
      trans = n.CatTransform(
        norm_mode='whatever',
        name='cat',
        index_to_cat_val=['a']
      )

    with self.assertRaises(TypeError):
      trans = n.CatTransform(
        norm_mode='min_max',
        name='cat',
      )

    trans = n.CatTransform(
      name='cat',
      norm_mode='mean_std',
      index_to_cat_val=['a']
    )
    with self.assertRaises(AssertionError):
      trans.pour(np.array([1]))


if __name__ == "__main__":
    unittest.main()



import shutil
import tempfile
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import wtrwrks.waterworks.waterwork as wa
from wtrwrks.waterworks.empty import empty
import numpy as np
import pprint
import os

class TestWaterwork(unittest.TestCase):
  def setUp(self):
      self.temp_dir = tempfile.mkdtemp()

  def tearDown(self):
      shutil.rmtree(self.temp_dir)

  def test_pour_pump_non_eager(self):
    with wa.Waterwork() as ww:
      add0_tubes, add0_slots = empty + empty

      add1_tubes, add1_slots = add0_tubes['target'] + empty

      cl0_tubes, _ = td.clone(a=add0_tubes['smaller_size_array'])

      add2_tubes, _ = td.add(a=cl0_tubes['a'], b=add1_tubes['smaller_size_array'])

      cl0_tubes['b'].set_name('output_1')

    true_funnel_dict = {
      ('Add_0', 'a'): np.array([1, 2]),
      ('Add_0', 'b'): np.array([3, 4]),
      ('Add_1', 'b'): np.array([1, 2])
    }
    for _ in xrange(2):
      self.assertEqual([str(t) for t in ww._pour_tank_order()], ['Add_0', 'Add_1', 'Clone_0', 'Add_2'])

      tap_dict = ww.pour(true_funnel_dict, key_type='str')

      true_tap_dict = {
          "output_1": np.array([3, 4]),
          'Add_0/tubes/a_is_smaller': False,
          'Add_1/tubes/a_is_smaller': False,
          'Add_1/tubes/target': np.array([5, 8]),
          'Add_2/tubes/a_is_smaller': False,
          'Add_2/tubes/smaller_size_array': np.array([1, 2]),
          'Add_2/tubes/a_is_smaller': False,
          'Add_2/tubes/target': np.array([4, 6]),
      }

      self.assertEqual(set(tap_dict.keys()), set(true_tap_dict.keys()))
      for tap in tap_dict:
        th.assert_arrays_equal(self, tap_dict[tap], true_tap_dict[tap])

      self.assertEqual(ww._pump_tank_order(), [ww.tanks[k] for k in ['Add_2', 'Add_1', 'Clone_0', 'Add_0']])

      funnel_dict = ww.pump(true_tap_dict, key_type='tuple')

      self.assertEqual(sorted(funnel_dict.keys()), sorted(true_funnel_dict.keys()))
      for funnel in funnel_dict:
        th.assert_arrays_equal(self, funnel_dict[funnel], true_funnel_dict[funnel])

      ww.clear_vals()
      for d in [ww.slots, ww.tubes]:
        for key in d:
          self.assertEqual(d[key].get_val(), None)
      pickle_name = os.path.join(self.temp_dir, 'ww.pickle')

      ww.save_to_file(pickle_name)
      ww = wa.Waterwork(from_file=pickle_name)

  def test_pour_pump_eager(self):
    with wa.Waterwork() as ww:
      cl0_tubes, cl0_slots = td.clone(a=np.array([1, 2]))
      cl0_slots['a'].unplug()

      add0_tubes, add0_slots = cl0_tubes['a'] + np.array([3, 4])
      add0_slots['b'].unplug()

      add1_tubes, _ = add0_tubes['target'] + cl0_tubes['b']
      cl1_tubes, _ = td.clone(a=add0_tubes['smaller_size_array'])
      add2_tubes, _ = cl1_tubes['a'] * add1_tubes['smaller_size_array']

      add2_tubes['target'].set_name('answer')

    true_funnel_dict = {
      cl0_slots['a']: np.array([1, 2]),
      add0_slots['b']: np.array([3, 4])
    }

    # print [str(t) for t in ww._pour_tank_order()]
    # self.assertEqual(ww._pour_tank_order(), [cl0, add0, cl1, add1, add2])
    true_tap_dict = {
        cl1_tubes['b']: np.array([3, 4]),
        add1_tubes['target']: np.array([5, 8]),
        add2_tubes['smaller_size_array']: np.array([1, 2]),
        "answer": np.array([3, 8]),
    }
    temp_tap_dict = {}
    temp_tap_dict.update(true_tap_dict)
    temp_tap_dict[add2_tubes['target']] = temp_tap_dict['answer']
    del temp_tap_dict['answer']
    for tap in temp_tap_dict:
      th.assert_arrays_equal(self, tap.get_val(), temp_tap_dict[tap])

    # print [str(t) for t in ww._pump_tank_order()]
    # self.assertEqual(ww._pump_tank_order(), [add2, cl1, add1, add0, cl0])

    # print [str(k) for k in ww.taps]
    funnel_dict = ww.pump(true_tap_dict)

    self.assertEqual(sorted(funnel_dict.keys()), sorted(true_funnel_dict.keys()))
    for funnel in funnel_dict:
      th.assert_arrays_equal(self, funnel_dict[funnel], true_funnel_dict[funnel])

  def test_pour_pump_iter_tube(self):
    with wa.Waterwork() as ww:
      ls, ls_slots = td.iter_list(empty, 2)
      ds, ds_slots = td.iter_dict(ls[1], ['a', 'b'])

      add0, _ = ls[0] + ds['a']
      add1, _ = ds['b'] + add0['target']
    true_funnel_dict = {
      ls_slots['a']: [7, {'a': 4, 'b': 5}]
    }
    tap_dict = ww.pour(true_funnel_dict)

    true_tap_dict = {
      add0['smaller_size_array']: 4,
      add0['a_is_smaller']: False,
      add1['target']: 16,
      add1['smaller_size_array']: 11,
      add1['a_is_smaller']: False,
    }
    temp_tap_dict = {}
    temp_tap_dict.update(true_tap_dict)

    self.assertEqual(sorted(tap_dict.keys()), sorted(temp_tap_dict.keys()))
    for tap in tap_dict:
      th.assert_arrays_equal(self, tap_dict[tap], temp_tap_dict[tap])

    funnel_dict = ww.pump(true_tap_dict)

    self.assertEqual(sorted(funnel_dict.keys()), sorted(true_funnel_dict.keys()))
    for funnel in funnel_dict:
      self.assertEqual(funnel_dict[funnel], true_funnel_dict[funnel])

    ww.clear_vals()
    for d in [ww.slots, ww.tubes]:
      for key in d:
        self.assertEqual(d[key].get_val(), None)

  def test_plug(self):
    with wa.Waterwork() as ww:
      add0_tubes, add0_slots = empty + empty

      add0_slots['b'].set_plug(np.array([3, 4]))
      add0_tubes['a_is_smaller'].set_plug(False)

      add1_tubes, add1_slots = add0_tubes['target'] + empty

      add1_slots['b'].set_plug(
        lambda d: 0.5 * d['Add_0/slots/a'] + np.array([0.5, 1.0])
      )
      add1_tubes['a_is_smaller'].set_plug(
        lambda d: not d['output_1'].any()
      )

      cl0_tubes, _ = td.clone(a=add0_tubes['smaller_size_array'])

      add2_tubes, _ = td.add(a=cl0_tubes['a'], b=add1_tubes['smaller_size_array'])

      add2_tubes['target'].set_plug(
        lambda d: d['output_1'] + np.array([1, 2])
      )

      cl0_tubes['b'].set_name('output_1')

    true_funnel_dict = {
      ('Add_0', 'a'): np.array([1, 2])
    }
    for _ in xrange(2):
      self.assertEqual([str(t) for t in ww._pour_tank_order()], ['Add_0', 'Add_1', 'Clone_0', 'Add_2'])

      tap_dict = ww.pour(true_funnel_dict, key_type='str')

      true_tap_dict = {
          "output_1": np.array([3, 4]),
          'Add_1/tubes/target': np.array([5, 8]),
          'Add_2/tubes/a_is_smaller': False,
          'Add_2/tubes/smaller_size_array': np.array([1, 2]),
          'Add_2/tubes/a_is_smaller': False,
          # 'Add_2/tubes/target': np.array([4, 6]),
      }

      self.assertEqual(set(tap_dict.keys()), set(true_tap_dict.keys()))
      for tap in tap_dict:
        th.assert_arrays_equal(self, tap_dict[tap], true_tap_dict[tap])

      self.assertEqual(ww._pump_tank_order(), [ww.tanks[k] for k in ['Add_2', 'Add_1', 'Clone_0', 'Add_0']])

      funnel_dict = ww.pump(true_tap_dict, key_type='tuple')

      self.assertEqual(sorted(funnel_dict.keys()), sorted(true_funnel_dict.keys()))
      for funnel in funnel_dict:
        th.assert_arrays_equal(self, funnel_dict[funnel], true_funnel_dict[funnel])

      ww.clear_vals()
      for d in [ww.slots, ww.tubes]:
        for key in d:
          self.assertEqual(d[key].get_val(), None)
      pickle_name = os.path.join(self.temp_dir, 'ww.pickle')

      ww.save_to_file(pickle_name)
      ww = wa.Waterwork(from_file=pickle_name)

  def test_list(self):
    with wa.Waterwork() as ww:
      add0_tubes, add0_slots = empty + empty

      add0_slots['b'].set_plug(2)
      add0_tubes['a_is_smaller'].set_plug(False)

      reshape = td.reshape(empty, [1, add0_tubes['smaller_size_array']])
    true_funnel_dict = {
      ('Add_0', 'a'): np.array([1, 2]),
      ('Reshape_0', 'a'): np.array([[[3, 4]]])
    }
    for _ in xrange(2):
      tap_dict = ww.pour(true_funnel_dict, key_type='str')

      true_tap_dict = {
        'Reshape_0/tubes/target': np.array([[3, 4]]),
        'Reshape_0/tubes/old_shape': [1, 1, 2],
        'Add_0/tubes/target': np.array([3, 4]),
      }

      self.assertEqual(set(tap_dict.keys()), set(true_tap_dict.keys()))
      for tap in tap_dict:
        th.assert_arrays_equal(self, tap_dict[tap], true_tap_dict[tap])

      funnel_dict = ww.pump(true_tap_dict, key_type='tuple')

      self.assertEqual(sorted(funnel_dict.keys()), sorted(true_funnel_dict.keys()))
      for funnel in funnel_dict:
        th.assert_arrays_equal(self, funnel_dict[funnel], true_funnel_dict[funnel])

      ww.clear_vals()
      for d in [ww.slots, ww.tubes]:
        for key in d:
          self.assertEqual(d[key].get_val(), None)
      pickle_name = os.path.join(self.temp_dir, 'ww.pickle')

      ww.save_to_file(pickle_name)
      ww = wa.Waterwork(from_file=pickle_name)

  def test_multi(self):
    with wa.Waterwork() as ww:
      add0_tubes, add0_slots = empty + empty

      add0_slots['b'].set_plug(np.array([3, 4]))
      add0_tubes['a_is_smaller'].set_plug(False)

      add1_tubes, add1_slots = add0_tubes['target'] + empty

      add1_slots['b'].set_plug(
        lambda d: 0.5 * d['Add_0/slots/a'] + np.array([0.5, 1.0])
      )
      add1_tubes['a_is_smaller'].set_plug(
        lambda d: not d['output_1'].any()
      )

      cl0_tubes, _ = td.clone(a=add0_tubes['smaller_size_array'])

      add2_tubes, _ = td.add(a=cl0_tubes['a'], b=add1_tubes['smaller_size_array'])

      add2_tubes['target'].set_plug(
        lambda d: d['output_1'] + np.array([1, 2])
      )

      cl0_tubes['b'].set_name('output_1')

    true_funnel_dict = {
      ('Add_0', 'a'): np.array([1, 2])
    }
    funnel_dicts = [true_funnel_dict] * 3
    tap_dicts = ww.multi_pour(funnel_dicts, key_type='str')
    for tap_dict in tap_dicts:
      # tap_dict = ww.pour(true_funnel_dict, key_type='str')

      true_tap_dict = {
          "output_1": np.array([3, 4]),
          'Add_1/tubes/target': np.array([5, 8]),
          'Add_2/tubes/a_is_smaller': False,
          'Add_2/tubes/smaller_size_array': np.array([1, 2]),
          'Add_2/tubes/a_is_smaller': False,
          # 'Add_2/tubes/target': np.array([4, 6]),
      }

      self.assertEqual(set(tap_dict.keys()), set(true_tap_dict.keys()))
      for tap in tap_dict:
        th.assert_arrays_equal(self, tap_dict[tap], true_tap_dict[tap])

      self.assertEqual(ww._pump_tank_order(), [ww.tanks[k] for k in ['Add_2', 'Add_1', 'Clone_0', 'Add_0']])

    funnel_dicts = ww.multi_pump(tap_dicts, key_type='tuple')
    for funnel_dict in funnel_dicts:
      self.assertEqual(sorted(funnel_dict.keys()), sorted(true_funnel_dict.keys()))
      for funnel in funnel_dict:
        th.assert_arrays_equal(self, funnel_dict[funnel], true_funnel_dict[funnel])

      ww.clear_vals()
      for d in [ww.slots, ww.tubes]:
        for key in d:
          self.assertEqual(d[key].get_val(), None)
      pickle_name = os.path.join(self.temp_dir, 'ww.pickle')


if __name__ == "__main__":
    unittest.main()



"""WaterworkPart definition."""
import wtrwrks.waterworks.globs as gl
import os


class WaterworkPart(object):
  """Base class for the components of a Waterwork. Primarly used to set the waterwork the part belongs to as well as its name within that waterwork.

  Attributes
  ----------
  waterwork : Waterwork or None
    The waterwork that the part will be added to. If None, it is assinged to the _default_waterwork.
  name : str or None
    The name of the part within the waterwork. Must be unique. If None, it will be set to a default value depending on the subclass.

  """

  def __init__(self, waterwork, name):
    """Define a WaterworkPart, assign it to a Waterwork and set its name within the Waterwork.

    Parameters
    ----------
    waterwork : Waterwork or None
      The waterwork that the part will be added to. If None, it is assinged to the _default_waterwork.
    name : str or None
      The name of the part within the waterwork. Must be unique. If None, it will be set to a default value depending on the subclass.

    """
    # Set the waterwork to add the tank to
    self.waterwork = waterwork
    if waterwork is None and gl._default_waterwork is None:
      raise ValueError("Must define op within 'with' statement of waterwork or pass a waterwork as an argument.")
    elif waterwork is None:
      self.waterwork = gl._default_waterwork

    # Set the namespace the part was created in
    self.name_space = gl._name_space

    # Set the tank name. Check and make sure it's valid.
    self.name = name
    pre = '' if self.name_space is None else self.name_space._get_name_string()

    if name is None:
      self.name = self._get_default_name(prefix=pre)
    elif type(name) not in (str, unicode):
      raise TypeError("'name' must be of type str or unicode. Got " + str(type(name)))
    elif not self.name.startswith(pre):
      self.name = os.path.join(pre, self.name)

  def _get_default_name(self):
    """Set a default name. Must be defined by subclass."""
    pass



import shutil
import tempfile
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.transforms.num_transform as n
import os
import pandas as pd
import numpy as np

class TestNumTransform(th.TestTransform):
    def setUp(self):
        self.temp_dir = tempfile.mkdtemp()
        self.array = np.array([
          [1, 2, np.nan],
          [4, np.nan, np.nan],
          [7, 8, np.nan],
          [10, 11, np.nan]
        ])

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_no_norm(self):
        trans = n.NumTransform(
          name='num'
        )
        trans.calc_global_values(self.array[:, 0: 1])
        target = self.array[:, 0: 1]
        for i in xrange(2):
          self.pour_pump(
            trans,
            self.array[:, 0: 1],
            {
              'num/nums': target,
              'num/nans': [[False], [False], [False], [False]],
            }
          )
          trans = self.write_read(trans, self.temp_dir)

    def test_nan(self):
      def fill(array):
        mins = np.expand_dims(np.nanmin(array, axis=0), axis=0)
        mins = np.tile(mins, reps=[4, 1])
        replace_with = mins[np.isnan(array)]
        return replace_with
      trans = n.NumTransform(
        name='num',
        fill_nan_func=fill
      )
      trans.calc_global_values(self.array[:, 1: 2])
      target = np.array(self.array[:, 1: 2], copy=True)
      target[1, 0] = 2
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 1: 2],
          {
            'num/nums': target,
            'num/nans': [[False], [True], [False], [False]],
          }
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_mean_std(self):
      def fill(array):
        return np.array(0.)

      trans = n.NumTransform(
        name='num',
        norm_mode='mean_std',
        fill_nan_func=fill
      )
      trans.calc_global_values(self.array[:, 0: 1])
      target = self.array[:, 0: 1]
      target = (target - trans.mean)/(trans.std)
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'num/nums': target,
            'num/nans': [[False], [False], [False], [False]],
          }
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_min_max(self):
      def fill(array):
        return np.array(0.0)
      trans = n.NumTransform(
        name='num',
        norm_mode='min_max',
        fill_nan_func=fill,
        norm_axis=0
      )
      trans.calc_global_values(self.array[:, 0: 2])
      target = self.array[:, 0: 2]
      target = (target - trans.min)/(trans.max - trans.min)
      target[1, 1] = -trans.min[1]/(trans.max[1] - trans.min[1])
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 2],
          {
            'num/nums': target,
            'num/nans': [[False, False], [False, True], [False, False], [False, False]],
          }
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_df(self):
      def fill(array):
        return np.array(0.)

      trans = n.NumTransform(
        name='num',
        norm_mode='mean_std',
        fill_nan_func=fill
      )
      df = pd.DataFrame({'a': self.array[0]})
      # trans.calc_global_values(self.array[:, 0: 1])
      trans.calc_global_values(data=df)
      target = self.array[:, 0: 1]
      target = (target - trans.mean)/(trans.std)
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'num/nums': target,
            'num/nans': [[False], [False], [False], [False]],
          }
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_array_iter(self):
      def fill(array):
        return np.array(0.0)
      trans = n.NumTransform(
        name='num',
        norm_mode='min_max',
        fill_nan_func=fill,
        norm_axis=0
      )
      array_iter = [self.array[0: 2, 0: 2], self.array[2: 4, 0: 2]]
      trans.calc_global_values(data_iter=array_iter)
      # trans.calc_global_values(self.array[:, 0: 2])
      target = self.array[:, 0: 2]
      target = (target - trans.min)/(trans.max - trans.min)
      target[1, 1] = -trans.min[1]/(trans.max[1] - trans.min[1])
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 2],
          {
            'num/nums': target,
            'num/nans': [[False, False], [False, True], [False, False], [False, False]],
          }
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_read_write(self):
      def fill(array):
        return np.array(0.0)
      trans = n.NumTransform(
        name='NUM',
        norm_mode='min_max',

        fill_nan_func=lambda a: np.array(0),
      )
      trans.calc_global_values(self.array)
      for i in xrange(2):
        self.write_read_example(trans, self.array, self.temp_dir, test_type=False)

    def test_errors(self):

      with self.assertRaises(ValueError):
        trans = n.NumTransform(
          norm_mode='whatever',
          name='num'
        )

      trans = n.NumTransform(
        name='num',
        norm_mode='min_max',
      )

      with self.assertRaises(AssertionError):
        trans.get_waterwork()



if __name__ == "__main__":
    unittest.main()



"""LowerCase tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import wtrwrks.string_manipulations.diff as di
import numpy as np
# from chop.mmseg import Tokenizer as MMSEGTokenizer
# from chop.hmm import Tokenizer as HMMTokenizer
import nltk
import tinysegmenter


class LowerCase(ta.Tank):
  """The defintion of the LowerCase tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'lower_case'
  slot_keys = ['strings']
  tube_keys = ['target', 'diff']

  def _pour(self, strings):
    """Execute the LowerCase tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of strings
      The array of strings to lower case

    Returns
    -------
    dict(
      target: np.ndarray of strings
        The array of lower cased strings.
      diff: np.ndarray of strings
        The string difference between the original strings and the lower cased strings.
    )

    """
    strings = np.array(strings)
    target = np.char.lower(strings)

    diff = np.vectorize(di.get_diff_string)(target, strings)
    self.target = target
    self.diff = diff
    return {'target': target, 'diff': diff}

  def _pump(self, target, diff):
    """Execute the LowerCase tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray of strings
      The array of lower cased strings.
    diff: np.ndarray of strings
      The string difference between the original strings and the lower cased strings.

    Returns
    -------
    dict(
      strings: np.ndarray of strings
        The array of strings to lower case
    )

    """

    strings = np.vectorize(di.reconstruct)(target, diff)

    return {'strings': strings}



import shutil
import tempfile
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.transforms.datetime_transform as n
import os
import pandas as pd
import numpy as np
import datetime


class TestDateTimeTransform(th.TestTransform):
    def setUp(self):
      self.temp_dir = tempfile.mkdtemp()
      self.array = np.array([
        ['2019-01-01', '2019-01-01', np.datetime64('NaT')],
        ['2019-01-02', np.datetime64('NaT'), np.datetime64('NaT')],
        ['2019-01-03', '2019-02-01', np.datetime64('NaT')],
        ['2019-01-01', '2019-03-01', np.datetime64('NaT')]
      ], dtype=np.datetime64)

    def tearDown(self):
      shutil.rmtree(self.temp_dir)

    def test_no_norm(self):
      def fill(array):
        return np.array(datetime.datetime(1970, 1, 1))
      trans = n.DateTimeTransform(
        name='datetime',
        # fill_nat_func=fill
      )
      trans.calc_global_values(self.array[:, 0: 1])
      target = np.array((self.array[:, 0: 1] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [False], [False], [False]],
            'datetime/diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_nan(self):
      def fill(array):
        return np.array(datetime.datetime(1970, 1, 1))
      trans = n.DateTimeTransform(
        name='datetime',
        fill_nat_func=fill,
        time_unit='W',
        num_units=2,
        zero_datetime=datetime.datetime(2000, 1, 1)
      )
      trans.calc_global_values(self.array[:, 1: 2])
      target = np.array((self.array[:, 1: 2] - trans.zero_datetime) / np.timedelta64(2, 'W'), copy=True)
      target[np.isnan(target)] = -782.64285714
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 1: 2],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [True], [False], [False]],
            'datetime/diff': np.array([[259200000000], [-172800000000], [518400000000], [518400000000]], dtype='timedelta64[us]')
          },
          test_type=False
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_mean_std(self):

      def fill(array):
        unique = np.unique(array[~np.isnat(array)])
        replace_with = np.full(array[np.isnat(array)].shape, unique[0])
        return replace_with

      trans = n.DateTimeTransform(
        name='',
        norm_mode='mean_std',
        fill_nat_func=fill
      )
      trans.calc_global_values(self.array[:, 1: 2])
      target = np.array((self.array[:, 1: 2] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      target = np.array([[-0.9153226063677012], [-0.9153226063677012], [0.34578854018335375], [1.4848566725520487]])
      # target[1, 0] = -1.24496691
      print target
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 1: 2],
          {
            'nums': target,
            'nats': [[False], [True], [False], [False]],
            'diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_mean_std_norm_axis(self):

      def fill(array):
        unique = np.unique(array[~np.isnat(array)])
        replace_with = np.full(array[np.isnat(array)].shape, unique[0])
        return replace_with

      trans = n.DateTimeTransform(
        name='datetime',
        norm_mode='mean_std',
        norm_axis=None,
        fill_nat_func=fill
      )
      trans.calc_global_values(self.array[:, 1: 2])
      target = np.array((self.array[:, 1: 2] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      target = np.array([[-0.9153226063677012], [-0.9153226063677012], [0.34578854018335375], [1.4848566725520487]])
      # target[1, 0] = -1.24496691
      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 1: 2],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [True], [False], [False]],
            'datetime/diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False
        )
        trans = self.write_read(trans, self.temp_dir)

    def test_min_max(self):
      def fill(array):
        mins = np.expand_dims(np.min(array, axis=0), axis=0)
        mins = np.tile(mins, reps=[4, 1])
        replace_with = mins[np.isnat(array)]
        return replace_with
      trans = n.DateTimeTransform(
        name='datetime',
        fill_nat_func=fill,
        norm_mode='min_max'
      )
      trans.calc_global_values(self.array[:, 0: 1])
      target = np.array((self.array[:, 0: 1] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      target = (target - trans.min)/(trans.max - trans.min)

      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [False], [False], [False]],
            'datetime/diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False

        )
        trans = self.write_read(trans, self.temp_dir)

    def test_df(self):
      def fill(array):
        mins = np.expand_dims(np.min(array, axis=0), axis=0)
        mins = np.tile(mins, reps=[array.shape[0], 1])
        replace_with = mins[np.isnat(array)]
        return replace_with
      trans = n.DateTimeTransform(
        name='datetime',
        fill_nat_func=fill,
        norm_mode='min_max'
      )
      df = pd.DataFrame({'a': self.array[0]})
      # trans.calc_global_values(self.array[:, 0: 1])
      trans.calc_global_values(data=df)
      target = np.array((self.array[:, 0: 1] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      target = (target - trans.min)/(trans.max - trans.min)

      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [False], [False], [False]],
            'datetime/diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False

        )
        trans = self.write_read(trans, self.temp_dir)

    def test_array_iter(self):
      def fill(array):
        mins = np.expand_dims(np.min(array, axis=0), axis=0)
        mins = np.tile(mins, reps=[array.shape[0], 1])
        replace_with = mins[np.isnat(array)]
        return replace_with
      trans = n.DateTimeTransform(
        name='datetime',
        fill_nat_func=fill,
        norm_mode='min_max'
      )
      array_iter = [self.array[0: 2, 0: 1], self.array[2: 4, 0: 1]]
      trans.calc_global_values(data_iter=array_iter)
      # trans.calc_global_values(self.array[:, 0: 1])
      target = np.array((self.array[:, 0: 1] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      target = (target - trans.min)/(trans.max - trans.min)

      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [False], [False], [False]],
            'datetime/diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False

        )
        trans = self.write_read(trans, self.temp_dir)

    def test_read_write(self):
      def fill(array):
        mins = np.expand_dims(np.min(array, axis=0), axis=0)
        mins = np.tile(mins, reps=[4, 1])
        replace_with = mins[np.isnat(array)]
        return replace_with
      trans = n.DateTimeTransform(
        name='datetime',
        fill_nat_func=fill,
        norm_mode='min_max'
      )
      trans.calc_global_values(self.array[:, 0: 1])
      target = np.array((self.array[:, 0: 1] - trans.zero_datetime) / np.timedelta64(1, 'D'), copy=True)
      target = (target - trans.min)/(trans.max - trans.min)

      for i in xrange(2):
        self.pour_pump(
          trans,
          self.array[:, 0: 1],
          {
            'datetime/nums': target,
            'datetime/nats': [[False], [False], [False], [False]],
            'datetime/diff': np.array([[datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)], [datetime.timedelta(0)]], dtype='timedelta64[us]')
          },
          test_type=False
        )
        self.write_read_example(trans, self.array[:, 0: 1].astype(np.datetime64), self.temp_dir, test_type=False)
        trans = self.write_read(trans, self.temp_dir)

    def test_errors(self):

      with self.assertRaises(ValueError):
        trans = n.DateTimeTransform(
          norm_mode='whatever',
          name='datetime'
        )

      with self.assertRaises(ValueError):
        trans = n.DateTimeTransform(
          name='datetime',
          norm_mode='whatever',
        )

      trans = n.DateTimeTransform(
        name='datetime',
        norm_mode='mean_std',
      )
      with self.assertRaises(AssertionError):
        trans.get_waterwork()


if __name__ == "__main__":
    unittest.main()



"""Definition of the StringTransform."""
import transform as n
import numpy as np
import wtrwrks.tanks.tank_defs as td
import wtrwrks.read_write.tf_features as feat
from wtrwrks.waterworks.empty import empty
import tensorflow as tf
import operator
import unicodedata

def _half_width(string):
  return unicodedata.normalize('NFKC', unicode(string))


class MultiLingualStringTransform(n.Transform):
  """Object that transforms raw strings into vectorized data with all the necessary data needed in order to completely reconstruct the original strings. Is used to handle several different languages at once.

  Parameters
  ----------
  name : str
    The name of the transform.
  dtype : numpy dtype
    The data type the transformed data should have. Defaults to np.float64.
  input_dtype: numpy dtype
    The datatype of the original inputted array.
  lower_case : bool
    Whether or not to lower_case all the strings.
  half_width : bool
    Whether or not to convert all full width characters to half width.
  index_to_word_maps : dic of lists
    A dictionary which maps a language to a list. Each list is a mapping from index number to word.
  max_sent_len : int
    The maximum allowed number of words in a sentence. Also decides the inner most dimension of the outputted indices array.
  max_vocab_size : int
    The maximum allowed number of words in the vocabulary.
  word_tokenizers : dict of funcs
    A dicitonary of which maps a language to a function. The functions take in a string and split it up into a list of words.
  word_detokenizers : dict of funcs
    A dicitonary of which maps a language to a function. The functions take in a list of words and output a string. Doesn't have to be an exact inverse to word_tokenizer but should be close otherwise a lot of large diff strings will have to be outputted in order to reproduce the original strings.

  Attributes
  ----------
  attribute_dict: dict
    The keys are the attributes of the class while the values are the default values. It's done this way rather than defined in the __init__ because this dictionary also defines what values need to be saved when written to disk, and what values need to be displayed when printed to the terminal.
  required_params: set of strs
    The parameters that must be provided to the transform at definition.
  cols : list of strs
    The column names of the array or dataframe to be transformed.
  num_examples : int
    The number of rows that have passed through the calc_global_values function.
  is_calc_run : bool
    Whether or not calc_global_values has been run for this transform.
  word_to_index_maps : dict of dict
    A dicitionary which maps a language to a another mapping. The mapping is from word to index number.

  """

  attribute_dict = {'name': '', 'dtype': np.int64, 'input_shape': None, 'index_to_word_maps': None, 'word_to_index_maps': None, 'max_sent_len': None, 'word_tokenizers': None, 'lemmatize': False, 'lemmatizer': None, 'half_width': False, 'lower_case': False, 'word_detokenizers': None, 'max_vocab_size': None}

  for k, v in n.Transform.attribute_dict.iteritems():
    if k in attribute_dict:
      continue
    attribute_dict[k] = v

  required_params = set(['word_tokenizers', 'word_detokenizers', 'max_sent_len'])
  required_params.update(n.Transform.required_params)

  def __init__(self, from_file=None, save_dict=None, **kwargs):
    """Define a transform using a dictionary, file, or by setting the attribute values in kwargs.

    Parameters
    ----------
    from_file : None or str
      The file path of the Tranform that was written to disk.
    save_dict : dict or None
      The dictionary of attributes that completely define a Transform.
    name : str
      The name of the transform.
    dtype : numpy dtype
      The data type the transformed data should have. Defaults to np.float64.
    input_dtype: numpy dtype
      The datatype of the original inputted array.
    lower_case : bool
      Whether or not to lower_case all the strings.
    half_width : bool
      Whether or not to convert all full width characters to half width.
    index_to_word_maps : dic of lists
      A dictionary which maps a language to a list. Each list is a mapping from index number to word.
    max_sent_len : int
      The maximum allowed number of words in a sentence. Also decides the inner most dimension of the outputted indices array.
    max_vocab_size : int
      The maximum allowed number of words in the vocabulary.
    word_tokenizers : dict of funcs
      A dicitonary of which maps a language to a function. The functions take in a string and split it up into a list of words.
    word_detokenizers : dict of funcs
      A dicitonary of which maps a language to a function. The functions take in a list of words and output a string. Doesn't have to be an exact inverse to word_tokenizer but should be close otherwise a lot of large diff strings will have to be outputted in order to reproduce the original strings.
    **kwargs :
      The keyword arguments that set the values of the attributes defined in the attribute_dict.

    """
    super(MultiLingualStringTransform, self).__init__(from_file, save_dict, **kwargs)

    # Require either a index to word mappings or a max vocab size if they are
    # to be built from scratch
    if self.index_to_word_maps is None and self.max_vocab_size is None:
      raise ValueError("Must supply index_to_word_maps mapping or a max_vocab_size.")

  def __len__(self):
    """Get the length of the transformed data."""
    return self.max_sent_len

  def _calc_global_values(self, array):
    """Calculate all the values of the Transform that are dependent on all the examples of the dataset. (e.g. mean, standard deviation, unique category values, etc.) This method must be run before any actual transformation can be done.

    Parameters
    ----------
    array : np.ndarray
      Some of the data that will be transformed.

    """
    if self.dtype is None:
      self.dtype = array.dtype
    else:
      array = np.array(array, dtype=self.input_dtype)
    if len(array.shape) < 2 or array.shape[1] != 2:
      raise ValueError("Array must have exactly two columns. The first being the string, and the second being the language.")

    languages = np.unique(array[:, 1])
    for language in languages:
      if language not in self.word_tokenizers or language not in self.word_detokenizers:
        raise ValueError("All languages must appear in both word_tokenizers and word_detokenizers. {} not found".format(language))

    if self.index_to_word_maps is None:
      for language in sorted(self.word_tokenizers):
        mask = array[:, 1: 2] == language
        lang_array = array[:, 0: 1][mask]

        # Get all the words and the number of times they appear
        strings = lang_array.flatten()

        if strings.size == 0:
          continue

        if self.lower_case:
          strings = np.char.lower(strings)

        if self.half_width:
          strings = np.vectorize(_half_width)(strings)

        word_tokenizer = self.word_tokenizers[language]

        for string in strings:
          # Tokenize each request, add the tokens to the set of all words
          tokens = word_tokenizer(string)
          for token_num, token in enumerate(tokens):
            self.all_words[language].setdefault(token, 0)
            self.all_words[language][token] += 1

    else:
      self.max_vocab_size = max([len(v) for v in self.index_to_word_maps.items()])

  def _finish_calc(self):
    """Finish up the calc global value process."""
    if self.index_to_word_maps is None:
      self.index_to_word_maps = {}
      for language in sorted(self.word_tokenizers):
        # Sort the dict by the number of times the words appear
        sorted_words = sorted(self.all_words[language].items(), key=operator.itemgetter(1), reverse=True)

        # Pull out the first 'max_vocab_size' words
        corrected_vocab_size = self.max_vocab_size - 1
        sorted_words = [w for w, c in sorted_words[:corrected_vocab_size]]

        # Create the mapping from category values to index in the vector and
        # vice versa
        self.index_to_word_maps[language] = ['[UNK]'] + sorted(sorted_words)
    else:
      self.max_vocab_size = max([len(self.index_to_word_maps[k]) for k in self.word_tokenizers])
      for language in sorted(self.word_tokenizers):
        token = self.index_to_word_maps[language][0]
        if token != '[UNK]':
          raise ValueError("First element of the every index_to_word_map must be the '[UNK]' token. Got {} for {}".format(token, language))

    self.word_to_index_maps = {}
    for language in sorted(self.word_tokenizers):
      self.word_to_index_maps[language] = {
        word: num for num, word in enumerate(self.index_to_word_maps[language])
      }

  def _get_array_attributes(self, prefix=''):
    """Get the dictionary that contain the original shapes of the arrays before being converted into tfrecord examples.

    Parameters
    ----------
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    array_attributes : dict
      The dictionary with keys equal to those that are found in the Transform's example dicts and values are the shapes of the arrays of a single example.

    """
    # Create the list of all the keys expected from the example_dicts
    array_keys = self._get_array_keys()

    # Get the original array's shape, except for the batch dim.
    shape = [1]

    att_dict = {}
    for key in array_keys:

      # Add a max_sent_len dim to the tokenize_diff array since it has a
      # diff for each token, otherwise give it input array's shape.
      cur_shape = shape if key in ('tokenize_diff', 'languages') else shape + [self.max_sent_len]

      att_dict[key] = {
        'shape': list(cur_shape),
        'tf_type': tf.int64 if key == 'indices' else tf.string,
        'size': feat.size_from_shape(cur_shape),
        'feature_func': feat._int_feat if key == 'indices' else feat._bytes_feat,
        'np_type': np.int64 if key == 'indices' else np.unicode
      }
    att_dict = self._pre(att_dict, prefix)
    return att_dict

  def _get_array_keys(self):
    """Get the relevant list of keys for this instance of the string transform"""
    array_keys = ['indices', 'tokenize_diff', 'missing_vals', 'languages']
    if self.lower_case:
      array_keys.append('lower_case_diff')
    if self.half_width:
      array_keys.append('half_width_diff')
    if self.lemmatize:
      array_keys.append('lemmatize_diff')
    return array_keys

  def _save_dict(self):
    """Create the dictionary of values needed in order to reconstruct the transform."""
    save_dict = {}
    for key in self.attribute_dict:
      save_dict[key] = getattr(self, key)
    save_dict['__class__'] = str(self.__class__.__name__)
    return save_dict

  def _start_calc(self):
    """Start the calc global value process."""
    # Create the mapping from category values to index in the vector and
    # vice versa
    self.all_words = {l: {} for l in self.word_tokenizers}

  def define_waterwork(self, array=empty, return_tubes=None, prefix=''):
    """Get the waterwork that completely describes the pour and pump transformations.

    Parameters
    ----------
    array : np.ndarray or empty
      The array to be transformed.

    Returns
    -------
    Waterwork
      The waterwork with all the tanks (operations) added, and names set.

    """
    splits, splits_slots = td.split(array, [1], axis=1)
    splits_slots['a'].unplug()
    splits_slots['a'].set_name('array')

    splits, _ = td.iter_list(splits['target'], 2)

    # Tokenize the full strings into words
    tokens, tokens_slots = td.multi_tokenize(
      strings=splits[0],
      selector=splits[1],
      tokenizers=self.word_tokenizers,
      detokenizers=self.word_detokenizers,
      max_len=self.max_sent_len
    )

    # Set the names of various tubes and slots to make it easier to reference
    # them in further downstream.
    tokens['diff'].set_name('tokenize_diff')
    tokens_slots['max_len'].set_name('max_sent_len')
    tokens_slots['tokenizers'].set_name('tokenizers')
    tokens_slots['detokenizers'].set_name('detokenizers')

    # lower_case the strings, and set the diff strings of the tank to
    # 'lower_case_dff' for easier referencing.
    if self.lower_case:
      tokens, tokens_slots = td.lower_case(tokens['target'])
      tokens['diff'].set_name('lower_case_diff')

    # Half width the strings, and set the diff strings of the tank to
    # 'half_width_diff' for easier referencing.
    if self.half_width:
      tokens, tokens_slots = td.half_width(tokens['target'])
      tokens['diff'].set_name('half_width_diff')

    # Lemmatize the strings, and set the diff strings of the tank to
    # 'lemmatize_dff' for easier referencing.
    if self.lemmatize:
      tokens, tokens_slots = td.lemmatize(tokens['target'])
      tokens['diff'].set_name('lemmatize_diff')
      tokens_slots['lemmatizer'].set_name('lemmatizer')

    languages, _ = td.clone(splits[1])
    languages['b'].set_name('languages')

    dim_size, _ = td.dim_size(languages['a'], axis=0)
    shape, _ = td.tube_list(dim_size['target'], 1, 1)
    tile, _ = td.reshape(
      languages['a'], shape['target'],
      tube_plugs={
        'old_shape': lambda z: (z[self._pre('languages', prefix)].shape[0], 1)
      }
    )
    tile, _ = td.tile(
      tile['target'], (1, 1, self.max_sent_len),
      tube_plugs={
        'old_shape': lambda z: (z[self._pre('languages', prefix)].shape[0], 1, 1)
      }
    )

    # Find all the strings which are not in the list of known words and
    # replace them with the 'unknown token'.
    maps_with_empty_strings = {k: v + [''] for k, v in self.index_to_word_maps.iteritems()}
    isin, isin_slots = td.multi_isin(tokens['target'], maps_with_empty_strings, tile['target'])

    mask, _ = td.logical_not(isin['target'])
    tokens, _ = td.replace(
      isin['a'], mask['target'], '[UNK]',
      tube_plugs={
        'mask': lambda z: z[self._pre('indices', prefix)] == 0
      }
    )

    # Keep track values that were overwritten with a 'unknown token'
    tokens['replaced_vals'].set_name('missing_vals')
    isin_slots['bs'].set_name('index_to_word_maps')

    # Convert the tokens into indices.
    indices, indices_slots = td.multi_cat_to_index(
      tokens['target'], tile['target'], self.word_to_index_maps,
      tube_plugs={
        'selector': lambda z: np.tile(np.reshape(z[self._pre('languages')], (z[self._pre('languages')].shape[0], 1, 1)), (1, 1, self.max_sent_len)),
        'missing_vals': lambda z: np.full(z[self._pre('indices')].shape, '', dtype=np.unicode),
        'input_dtype': self.input_dtype
      }
    )

    # Set the names of the slots and tubes of this tank for easier referencing
    indices['target'].set_name('indices')
    # indices['selector'].set_name('languages')
    indices_slots['cat_to_index_maps'].set_name('word_to_index_maps')

    if return_tubes is not None:
      ww = indices['target'].waterwork
      r_tubes = []
      for r_tube_key in return_tubes:
        r_tubes.append(ww.maybe_get_tube(r_tube_key))
      return r_tubes

  def get_schema_dict(self, var_lim=None):
    """Create a dictionary which defines the proper fields which are needed to store the untransformed data in a (postgres) SQL database.

    Parameters
    ----------
    var_lim : int or dict
      The maximum size of strings. If an int is passed then all VARCHAR fields have the same limit. If a dict is passed then each field gets it's own limit. Defaults to 255 for all string fields.

    Returns
    -------
    schema_dict : dict
      Dictionary where the keys are the field names and the values are the SQL data types.

    """
    lang_var_lim = 1
    for lang in self.word_tokenizers:
      if len(lang) > lang_var_lim:
        lang_var_lim = len(lang)
    if var_lim is None:
      var_lim = self.max_sent_len
      for lang in self.index_to_word_maps:
        index_to_word = self.index_to_word_maps[lang]
        for word in index_to_word:
          if len(word) * self.max_sent_len > var_lim:
            var_lim = len(word) * self.max_sent_len

    schema_dict = {self.cols[0]: 'VARCHAR({})'.format(var_lim), self.cols[1]: 'VARCHAR({})'.format(lang_var_lim)}
    return schema_dict



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestBools(th.TestTank):

  def test_isnan(self):
    a = np.array([[0, np.nan], [2, 3], [4, np.nan], [1, 0]])
    self.pour_pump(
      td.isnan,
      {
        'a': a,
      },
      {
        'a': a,
        'target': np.isnan(a),
      },
      type_dict={'a': np.ndarray}
    )

    a = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    a[0, 0, 0] = np.nan
    a[1, 0, 0] = np.nan
    a[0, 1, 0] = np.nan
    a[0, 0, 1] = np.nan
    a[1, 1, 1] = np.nan

    self.pour_pump(
      td.isnan,
      {
        'a': a,
      },
      {
        'target': np.isnan(a),
        'a': a,
      },
      type_dict={'a': np.ndarray}
    )

  def test_equal(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.equals,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    a = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    b = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.equals,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    b[0, 0, 1] = b[0, 0, 1] + 1
    target[0, 0, 1] = False
    self.pour_pump(
      td.equals,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

  def test_greater(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]]) - 1
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.greater,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    a = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    b = np.arange(24).reshape((4, 3, 2)).astype(np.float64) - 1
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.greater,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    b[0, 0, 1] = b[0, 0, 1] + 1
    target[0, 0, 1] = False
    self.pour_pump(
      td.greater,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

  def test_greater_equal(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.greater_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]]) + 1
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.greater_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    a = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    b = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.greater_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    b[0, 0, 1] = b[0, 0, 1] + 1
    target[0, 0, 1] = False
    self.pour_pump(
      td.greater_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

  def test_less(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]]) + 1
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.less,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    a = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    b = np.arange(24).reshape((4, 3, 2)).astype(np.float64) + 1
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.less,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    b[0, 0, 1] = b[0, 0, 1] - 1
    target[0, 0, 1] = False
    self.pour_pump(
      td.less,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

  def test_less_equal(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.less_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]]) - 1
    b = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.less_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    a = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    b = np.arange(24).reshape((4, 3, 2)).astype(np.float64)
    target = np.ones(a.shape).astype(bool)
    self.pour_pump(
      td.less_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    b[0, 0, 1] = b[0, 0, 1] - 1
    target[0, 0, 1] = False
    self.pour_pump(
      td.less_equal,
      {'a': a, 'b': b},
      {'a': a, 'b': b, 'target': target},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
if __name__ == "__main__":
    unittest.main()



import numpy as np


def empty_array_like(a, default_val=None):
  """Create an array of the same shape and type as 'a', with all default values.

  Parameters
  ----------
  a : np.ndarray of bools
    The array shape and dtype to copy


  Returns
  -------
  np.ndarray
    An array of all default values.

  """
  dtype = a.dtype

  # Choose different default values, and type for the returned array depending
  # on the input_dtype
  if dtype.type in (np.string_, np.unicode_):
    if default_val is None:
      default_val = ''
  elif dtype in (np.int64, np.int32, np.float64, np.float32):
    if default_val is None:
      default_val = 0
  elif dtype.type == np.timedelta64:
    default_val = 0
  elif dtype.type == np.datetime64:
    default_val = '1970-01-01'
  else:
    raise TypeError("Only string and number types are supported. Got " + str(dtype))

  full_array = np.full(a.shape, default_val, dtype=dtype)
  return full_array


def empty_array(shape, dtype, default_val=None):
  """Create an array of the same shape and type as 'a', with all default values.

  Parameters
  ----------
  a : np.ndarray of bools
    The array shape and dtype to copy


  Returns
  -------
  np.ndarray
    An array of all default values.

  """
  # Choose different default values, and type for the returned array depending
  # on the input_dtype
  if dtype.type in (np.string_, np.unicode_):
    if default_val is None:
      default_val = ''
  elif dtype in (np.int64, np.int32, np.float64, np.float32):
    if default_val is None:
      default_val = 0
  elif dtype.type == np.timedelta64:
    default_val = 0
  elif dtype.type == np.datetime64:
    default_val = '1970-01-01'
  elif dtype.type == np.dtype('O'):
    pass
  else:
    raise TypeError("Only string and number types are supported. Got " + str(dtype))

  full_array = np.full(shape, default_val, dtype=dtype)
  return full_array


def get_default_val(a):
  dtype = a.dtype
  # Choose different default values, and type for the returned array depending
  # on the input_dtype
  if dtype.type in (np.string_, np.unicode_):
    default_val = ''
  elif dtype in (np.int64, np.int32, np.float64, np.float32, np.timedelta64):
    default_val = 0
  elif dtype.type == np.datetime64:
    default_val = '1970-01-01'
  else:
    default_val = None

  return default_val



"""Div tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Div(ta.Tank):
  """The defintion of the Div tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'div'
  slot_keys = ['a', 'b']
  tube_keys = ['target', 'smaller_size_array', 'a_is_smaller', 'missing_vals', 'remainder']

  def _pour(self, a, b):
    """Execute the Div tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The numerator array.
    b: np.ndarray
      The denominator array

    Returns
    -------
    dict(
      target: np.ndarray
        The result of a/b
      smaller_size_array: np.ndarray
        Either 'a' or 'b' depending on which has fewer elements.
      a_is_smaller: bool
        Whether a is the smaller sized array.
      missing_vals: np.ndarray
        The values from either 'a' or 'b' that were lost when the other array had a zero in that location.
      remainder: np.ndarray
        The remainder of a/b in the case that 'a' and 'b' are of integer type.
    )

    """
    # If they aren't numpy arrays then cast them to arrays.
    if type(a) is not np.ndarray:
      a = np.array(a)
    if type(b) is not np.ndarray:
      b = np.array(b)

    # Find the array with fewer elements and save that.
    a_is_smaller = a.size < b.size
    if a_is_smaller:
      smaller_size_array = ut.maybe_copy(a)
    else:
      smaller_size_array = ut.maybe_copy(b)

    # Do the division
    target = np.array(a / b)

    # Save the values of the larger array whose values are erased by a zero in
    # the smaller array
    if a_is_smaller:
      missing_vals = b[(target == 0)]
    else:
      missing_vals = a[np.isposinf(target) | np.isneginf(target) | np.isnan(target)]

    # Don't allowed integer division by zero.
    if a.dtype in (np.int32, np.int64) and b.dtype in (np.int32, np.int64):
      if (b == 0).any():
        raise ZeroDivisionError("Integer division by zero is not supported.")
      remainder = np.array(np.remainder(a, b))
    else:
      remainder = np.array([], dtype=target.dtype)

    return {'target': target, 'smaller_size_array': smaller_size_array, 'a_is_smaller': a_is_smaller, 'missing_vals': missing_vals, 'remainder': remainder}

  def _pump(self, target, smaller_size_array, a_is_smaller, missing_vals, remainder):
    """Execute the Div tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The result of a/b
    smaller_size_array: np.ndarray
      Either 'a' or 'b' depending on which has fewer elements.
    a_is_smaller: bool
      Whether a is the smaller sized array.
    missing_vals: np.ndarray
      The values from either 'a' or 'b' that were lost when the other array had a zero in that location.
    remainder: np.ndarray
      The remainder of a/b in the case that 'a' and 'b' are of integer type.

    Returns
    -------
    dict(
      a: np.ndarray
        The numerator array.
      b: np.ndarray
        The denominator array
    )

    """
    if a_is_smaller:
      # If a is the smaller of the two arrays, then it was the one that was
      # saved. So no need to worry about the remainder.
      a = ut.maybe_copy(smaller_size_array)
      b = np.array(a / target)
      b[(target == 0)] = missing_vals
    else:
      a = target * smaller_size_array
      if target.dtype in (np.int32, np.int64):
        a = np.array(a + remainder)
      b = ut.maybe_copy(smaller_size_array)

      # If b is the smaller array then it is the one that was saved. This means
      # a nan, negative infinity, or positive infinity, (i.e. zeros in b)
      # correspond to the missing values in a.
      a[np.isposinf(target) | np.isneginf(target) | np.isnan(target)] = missing_vals
    return {'a': a, 'b': b}



"""Transpose tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np


class Transpose(ta.Tank):
  """The defintion of the Transpose tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

    Attributes
    ----------
    slot_keys: list of strs
      The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
    tubes: list of strs
      The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'transpose'
  slot_keys = ['a', 'axes']
  tube_keys = ['target', 'axes']
  pass_through_keys = ['axes']

  def _pour(self, a, axes):
    """Execute the Transpose tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to be transposed.
    axes: list of ints
      The permutation of axes. len(axes) must equal rank of 'a', and each integer from 0 to len(axes) - 1 must appear exactly once.

    Returns
    -------
    dict(
      target: np.ndarray
        The transposed array.
      axes: list of ints
        The permutation of axes. len(axes) must equal rank of 'a', and each integer from 0 to len(axes) - 1 must appear exactly once.
    )

    """
    return {'target': np.transpose(a, axes=axes), 'axes': axes}

  def _pump(self, target, axes):
    """Execute the Transpose tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The transposed array.
    axes: list of ints
      The permutation of axes. len(axes) must equal rank of 'a', and each integer from 0 to len(axes) - 1 must appear exactly once.

    Returns
    -------
    dict(
      a: np.ndarray
        The array to be transposed.
      axes: list of ints
        The permutation of axes. len(axes) must equal rank of 'a', and each integer from 0 to len(axes) - 1 must appear exactly once.
    )

    """
    # The inverse of a permutation is simply the argsort of that permutation.
    trans_axes = np.argsort(axes)
    a = np.transpose(target, axes=trans_axes)
    return {'a': a, 'axes': axes}



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestSum(th.TestTank):

  def test_one_d(self):
    self.pour_pump(
      td.sum,
      {'a': np.array([1, 3]), 'axis': ()},
      {'target': np.array(4), 'a': np.array([1, 3]), 'axis': ()},
      type_dict={'a': np.ndarray, 'axis': int},
      test_type=False
    )

  def test_two_d(self):
    self.pour_pump(
      td.sum,
      {'a': np.array([[0, 1], [2, 3], [4, 5], [1, 0]]), 'axis': 1},
      {
        'target': np.array([1, 5, 9, 1]),
        'a': np.array([[0, 1], [2, 3], [4, 5], [1, 0]]),
        'axis': 1
      },
      type_dict={'a': np.ndarray, 'axis': int}
    )

    self.pour_pump(
      td.sum,
      {'a': np.array([[0, 1], [2, 3], [4, 5], [1, 0]]), 'axis': 0},
      {
        'target': np.array([7, 9]),
        'a': np.array([[0, 1], [2, 3], [4, 5], [1, 0]]),
        'axis': 0
      },
      type_dict={'a': np.ndarray, 'axis': int}
    )

  def test_three_d(self):
    self.pour_pump(
      td.sum,
      {'a': np.arange(24).reshape((2, 3, 4)), 'axis': (0, 1)},
      {
        'target': np.array([60, 66, 72, 78]),
        'a': np.arange(24).reshape((2, 3, 4)),
        'axis': (0, 1)
      },
      type_dict={'a': np.ndarray, 'axis': int}
    )

if __name__ == "__main__":
    unittest.main()



"""
add_tensors : dict of tensors
  Any additional tensors to add into the dataset object.

batch_size : int
  The number of example to include in a single batch.

cat_val_to_index : dict
  The mapping from category value to index number.

cols : list of strs
  The column names of the array or dataframe to be transformed.

data : np.ndarray
  The numpy array to transform.

data : np.array or pd.DataFrame
  The entire dataset in the form of a numpy array or a pandas DataFrame. Should have the same columns as the arrays that will be fed to the pour method.

data_iter : iterator of np.array or pd.DataFrame
  The entire dataset in the form of an iterator of numpy array or a pandas DataFrame. Needed if the dataset is too large to fit in memory. Should have the same columns as the arrays that will be fed to the pour method. Can only use if 'data' is not being used

dataset_iter : tf.data.Iterator
  The dataset object to feed into tensorflow training pipelines.

drop_features : list of strs
  The features to drop when reading from the tfrecords. Defaults to None. Can only use this or keep_features not both.

dtype : numpy dtype
  The data type the transformed data should have. Defaults to np.float64.

examples : list of dicts of features
  The example dictionaries which contain tf.train.Features.

from_file : str
  The path to the saved file to recreate the transform object that was saved to disk.

file_name : str
  The name of the tfrecord file to write to. An extra '_<num>' will be added to the name.

file_name_pattern : string
  The file name pattern of the tfrecord files. Supports '*', etc.

file_num_offset : int
  A number that controls what number will be appended to the file name (so that files aren't overwritten.)

filters : dict of functions
  Any pre preprocessing filters to put on the dataset. The keys are the filter names, the values are the filter themselves

funnel_dicts : list of dicts(
  keys - Slot objects or Placeholder objects. The 'funnels' (i.e. unconnected slots) of the waterwork.
  values - valid input data types
)
  The inputs to the waterwork's full pour functions. There is exactly one funnel_dict for every process.


index_to_cat_val : list
  The mapping from index number to category value.

index_to_word : list
  The mapping from index number to word.

index_to_word_maps : dic of lists
  A dictionary which maps a language to a list. Each list is a mapping from index number to word.

input_dtype: numpy dtype
  The datatype of the original inputted array.

input_shape: list of ints
  The shape of the original inputted array.

keep_features : list of strs
  The features to keep after reading in from the tfrecords. Defaults to all of them.

name : str
  The name of the transform.

norm_axis : 0, 1, or None
  In the case that 'norm_mode' has a non None value, what axis should be used for the normalization. None implies both 0 and 1.

norm_mode : "mean_std" or None
  How to normalize the data. Subtracting out the mean and dividing by the standard deviation or leaving as is.

num_examples : int
  The number of rows that have passed through the calc_global_values function.

num_steps : int
  Number of steps to run in the dataset before terminating. Cannot use when num_epochs is defined

num_threads : int
  The number of io threads to use. Defaults to 1, should probably not be more than 3.

mean : numpy array
  The stored mean values to be used to normalize data.

prefix : str
  Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

recreate : bool
  Whether or not to force the transform to create a new waterwork.

save_dict : dict
  The dictionary to recreate the transform object.

schema_dict : dict
  Dictionary where the keys are the field names and the values are the SQL data types.

shuffle_buffer_size : int
  How many examples to shuffle together.

skip_fails : bool
  Whether or not to skip any write failures without error. Defaults to false.

skip_keys : list of strs
  Any taps that should not be written to examples.

std : numpy array
  The stored standar deviation values to be used to normalize data.

tap_dict : dict (or iterator of dicts)
  The dictionary of transformed outputs as well as any additional information needed to completely reconstruct the original data. Returns an iterator of dicts if something is passed to 'data_iter' rather than the 'data' argument.

tap_dict: dict
  The dictionary all information needed to completely reconstruct the original data.

to_prefix : str or dict
  The string or dictionary to add to the prefix/self.name prefix to.

to_unprefix : str or dict
  The string or dictionary to strip the prefix/self.name  from.

var_lim : int or dict
  The maximum size of strings. If an int is passed then all VARCHAR fields have the same limit. If a dict is passed then each field gets it's own limit. Defaults to 255 for all string fields.

Waterwork
  The waterwork object that this transform creates.

word_detokenizer : func
  A function that takes in a list of words and outputs a string. Doesn't have to be an exact inverse to word_tokenizer but should be close otherwise a lot of large diff strings will have to be outputted in order to reproduce the original strings.

word_detokenizers : dict of funcs
  A dicitonary of which maps a language to a function. The functions take in a list of words and output a string. Doesn't have to be an exact inverse to word_tokenizer but should be close otherwise a lot of large diff strings will have to be outputted in order to reproduce the original strings.

word_to_index : dict
  The mapping from word to index number.

word_to_index_maps : dict of dict
  A dicitionary which maps a language to a another mapping. The mapping is from word to index number.

word_tokenizers : dict of funcs
  A dicitonary of which maps a language to a function. The functions take in a string and split it up into a list of words.

word_tokenizer : func
  A function that takes in a string and splits it up into a list of words.



"""



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestCasts(th.TestTank):

  def test_np(self):
    a = np.array([[0, 1], [2, 3], [4, 5], [1, 0]])
    target = a.astype(float)
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': np.float64},
      {'target': target, 'diff': np.zeros(a.shape, dtype=float), 'input_dtype': a.dtype},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    a = np.array([[0, 1.5]])
    target = a.astype(int)
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': np.int64},
      {'target': target, 'diff': np.array([[0, 0.5]]), 'input_dtype': a.dtype},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    a = np.array([[0, 1.5]])
    target = a.astype(bool)
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': np.bool},
      {'target': target, 'diff': np.array([[0, 0.5]]), 'input_dtype': a.dtype},
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

  def test_basic(self):
    a = 1
    target = 1.0
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': float},
      {'target': target, 'diff': np.zeros((), dtype=float), 'input_dtype': type(a)},
      type_dict={'a': int, 'dtype': type(int)},
      test_type=False
    )
    a = 1.5
    target = 1
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': int},
      {'target': target, 'diff': np.array(0.5), 'input_dtype': type(a)},
      type_dict={'a': int, 'dtype': type(int)},
      test_type=False
    )
    a = 0.5
    target = True
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': bool},
      {'target': target, 'diff': -0.5, 'input_dtype': type(a)},
      test_type=False
    )
    a = 0
    target = False
    self.pour_pump(
      td.cast,
      {'a': a, 'dtype': bool},
      {'target': target, 'diff': 0, 'input_dtype': type(a)},
      type_dict={'a': int, 'dtype': type(int)},
      test_type=False
    )
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np
import tinysegmenter

class TestReplaceSubstring(th.TestTank):
  def test_scalar(self):
    strings = np.array("It is what it is.")
    diff = np.array('[["i", 6, 6, "what"]]')
    old_substring = 'what'
    new_substring = ''
    self.pour_pump(
      td.replace_substring,
      {
        'strings': strings,
        'old_substring': old_substring,
        'new_substring': new_substring
      },
      {
        'target': np.char.replace(strings, old_substring, new_substring),
        'diff': diff,
        'old_substring': old_substring,
        'new_substring': new_substring
      },
      test_type=False
    )

  def test_one_d(self):
    strings = np.array([
      u'チラシ・勧誘印刷物の無断投函は一切お断り',
      u'すみませんが、もう一度どお願いします。'
    ])
    diff = ['[]', '[]']
    old_substring = u'ー'
    new_substring = 'bamboozle'
    self.pour_pump(
      td.replace_substring,
      {
        'strings': strings,
        'old_substring': old_substring,
        'new_substring': new_substring
      },
      {
        'target': np.char.replace(strings, old_substring, new_substring),
        'diff': diff,
        'old_substring': old_substring,
        'new_substring': new_substring
      },
      test_type=False
    )

  def test_two_d(self):
    strings = np.array([
      [
        "It is what it is.",
        "Whatever, Bob's mother has seen the world."
      ],
      [
        "The sun is not yellow, it's chicken. OK.",
        "Don't need a weatherman to know which way the wind blows."
      ]
    ])
    diff = [
      ['[["i", 3, 12, "is"], ["i", 21, 30, "is"]]', '[]'],
      ['[["i", 8, 17, "is"]]', '[]']
    ]
    old_substring = 'is'
    new_substring = 'bamboozle'
    self.pour_pump(
      td.replace_substring,
      {
        'strings': strings,
        'old_substring': old_substring,
        'new_substring': new_substring
      },
      {
        'target': np.char.replace(strings, old_substring, new_substring),
        'diff': diff,
        'old_substring': old_substring,
        'new_substring': new_substring
      },
      test_type=False
    )

if __name__ == "__main__":
    unittest.main()



"""Concatenate tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np
import wtrwrks.tanks.utils as ut
import logging


class PackWithRowMap(ta.Tank):
  """More efficiently pack in the data of an array by overwriting the default_val's. The array must have rank at least equal to 2 The last dimension will be packed so that fewer default vals appear, and the next to last dimension with be shortened, any other dimensions are left unchanged.
  e.g.

  default_val = 0
  a = np.array([
    [1, 1, 1, 0, 0, 0],
    [2, 2, 0, 0, 0, 0],
    [3, 3, 0, 3, 3, 0],
    [4, 0, 0, 0, 0, 0],
    [5, 5, 5, 5, 5, 5],
    [6, 6, 0, 0, 0, 0],
    [7, 7, 0, 0, 0, 0]
  ])

  target = np.array([
    [1, 1, 1, 2, 2, 0],
    [3, 3, 3, 3, 4, 0],
    [5, 5, 5, 5, 5, 5],
    [6, 6, 7, 7, 0, 0]
  ])

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'pack_with_row_map'
  slot_keys = ['a', 'default_val', 'row_map']
  tube_keys = ['target', 'row_map', 'default_val']
  pass_through_keys = ['default_val', 'row_map']

  def _pour(self, a, row_map, default_val):
    """

    Parameters
    ----------
    a: np.ndarray
      The array to pack
    default_val: np.ndarray.dtype
      The value that will be allowed to be overwritten in the packing process.
    row_map: np.ndarray
      A mapping from the new rows to the original rows.

    Returns
    -------
    dict(
      target: np.ndarray
        The packed version of the 'a' array. Has same dims except for the second to last dimension which is usually shorter.
      default_val: np.ndarray.dtype
        The value that will be allowed to be overwritten in the packing process.
      row_map: np.ndarray
        A mapping from the new rows to the original rows.
    )

    """
    logging.debug('%s', a.shape)
    a = np.array(a)

    a_row_dim = a.shape[-1]
    sliced_row_map = row_map[..., 0]

    target_inner_dims = list(sliced_row_map.shape[-2:])
    reshaped_a = a.reshape([-1, a_row_dim])
    reshaped_row_map = sliced_row_map.reshape([-1] + target_inner_dims)

    # Flatten any of the outer dimensions and work with two d arrays, since
    # those will be the dimensions which affect the packing.
    target = []
    for slice_a, slice_row_map, in zip(reshaped_a, reshaped_row_map):
      slice_target = slice_a[slice_row_map]
      slice_target[slice_row_map == -1] = default_val
      target.append(slice_target.tolist())

    target = np.stack(target)
    target = target.reshape(row_map.shape[:-1])

    return {'target': target, 'default_val': default_val, 'row_map': row_map}

  def _pump(self, target, default_val, row_map):
    """Execute the Concatenate tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The packed version of the 'a' array. Has same dims except for the second to last dimension which is usually shorter.
    default_val: np.ndarray.dtype
      The value that will be allowed to be overwritten in the packing process.
    row_map: np.ndarray
      A mapping from the new rows to the original rows.

    Returns
    -------
    dict(
      a: np.ndarray
        The array to pack
      default_val: np.ndarray.dtype
        The value that will be allowed to be overwritten in the packing process.
      row_map: np.ndarray
        A mapping from the new rows to the original rows.
    )

    """
    target = np.array(target)
    dtype = target.dtype
    row_dim = target.shape[-2]
    col_dim = target.shape[-1]

    a_row_dim = np.max(row_map[..., 0]) + 1

    reshaped_target = target.reshape([-1, row_dim, col_dim])

    sliced_row_map = row_map[..., 0]
    reshaped_row_map = sliced_row_map.reshape([-1, row_dim, col_dim])

    a = []
    for slice_target, slice_row_map, in zip(reshaped_target, reshaped_row_map):
      slice_a = np.full([a_row_dim], default_val, dtype=dtype).tolist()

      for target_row_num in xrange(slice_target.shape[0]):
        for col_num, a_row_num in enumerate(slice_row_map[target_row_num]):
          if a_row_num == -1:
            continue

          slice_a[a_row_num] = slice_target[target_row_num][col_num]
      a.append(slice_a)
    a = np.stack(a).reshape(list(target.shape[:-2]) + [a_row_dim])
    return {'a': a, 'default_val': default_val, 'row_map': row_map}



"""ReplaceSubstring tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.string_manipulations.diff as di
import numpy as np


class ReplaceSubstring(ta.Tank):
  """The defintion of the ReplaceSubstring tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'replace_substring'
  slot_keys = ['strings', 'old_substring', 'new_substring']
  tube_keys = ['target', 'old_substring', 'new_substring', 'diff']
  pass_through_keys = ['old_substring', 'new_substring']

  def _pour(self, strings, old_substring, new_substring):
    """Execute the ReplaceSubstring tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    strings: np.ndarray of strings
      The array of strings that will have it's substrings replaced.
    old_substring: str or unicode
      The substring to be replaced.
    new_substring: str or unicode
      The substring to replace with.

    Returns
    -------
    dict(
      target: np.ndarray of same type as 'a'
        The of array of strings with the substrings replaced.
      old_substring: str or unicode
        The substring to be replaced.
      new_substring: str or unicode
        The substring to replace with.
      diff: np.ndarray of strings
        The diff of the strings caused by converting all old_substrings to new_substrings and back.
    )

    """
    strings = np.array(strings)
    target = np.char.replace(strings, old_substring, new_substring)

    diff = np.vectorize(di.get_diff_string)(target, strings)

    return {'target': target, 'diff': diff, 'old_substring': old_substring, 'new_substring': new_substring}

  def _pump(self, target, diff, old_substring, new_substring):
    """Execute the ReplaceSubstring tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray of same type as 'a'
      The of array of strings with the substrings replaced.
    old_substring: str or unicode
      The substring to be replaced.
    new_substring: str or unicode
      The substring to replace with.
    diff: np.ndarray of strings
      The diff of the strings caused by converting all old_substrings to new_substrings and back.

    Returns
    -------
    dict(
      strings: np.ndarray of strings
        The array of strings that will have it's substrings replaced.
      old_substring: str or unicode
        The substring to be replaced.
      new_substring: str or unicode
        The substring to replace with.
    )

    """
    strings = np.vectorize(di.reconstruct)(target, diff)

    return {'strings': strings, 'old_substring': old_substring, 'new_substring': new_substring}



"""Split tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Split(ta.Tank):
  """The defintion of the Split tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'split'
  slot_keys = ['a', 'indices', 'axis']
  tube_keys = ['target', 'indices', 'axis']
  pass_through_keys = ['indices', 'axis']

  def _pour(self, a, indices, axis):
    """Execute the Split tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to split up.
    indices: np.ndarray
      The indices of the points to split up the array.
    axis: int
      The axis along which to split up the array.

    Returns
    -------
    dict(
      target: list of arrays
        The list of split up arrays.
      indices: np.ndarray
        The indices of the points to split up the array.
      axis: int
        The axis along which to split up the array.
    )

    """
    return {'target': np.split(a, indices, axis=axis), 'indices': indices, 'axis': axis}

  def _pump(self, target, indices, axis):
    """Execute the Split tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: list of arrays
      The list of split up arrays.
    indices: np.ndarray
      The indices of the points to split up the array.
    axis: int
      The axis along which to split up the array.

    Returns
    -------
    dict(
      a: np.ndarray
        The array to split up.
      indices: np.ndarray
        The indices of the points to split up the array.
      axis: int
        The axis along which to split up the array.
    )

    """
    a = np.concatenate(target, axis=axis)
    return {'a': a, 'indices': indices, 'axis': axis}



"""NumTransform definition."""
import transform as n
import numpy as np
import logging
import wtrwrks.tanks.tank_defs as td
import wtrwrks.read_write.tf_features as feat
import tensorflow as tf
from wtrwrks.waterworks.empty import empty


class NumTransform(n.Transform):
  """Class used to create mappings from raw numerical data to vectorized, normalized data and vice versa.

  Parameters
  ----------
  name : str
    The name of the transform.
  dtype : numpy dtype
    The data type the transformed data should have. Defaults to np.float64.
  input_dtype: numpy dtype
    The datatype of the original inputted array.
  norm_mode : "mean_std", "min_max" or None
    How to normalize the data. Subtracting out the mean and dividing by the standard deviation; Subtracting out min and dividing by the difference between max and min; or leaving as is.
  norm_axis : 0, 1, or None
    In the case that 'norm_mode' has a non None value, what axis should be used for the normalization. None implies both 0 and 1.
  fill_nan_func : np.array(dtype=np.float64) -> np.array(dtype=np.float64)
    Function which fills in NaN values with some value. This function takes in a numpy array which contains NaN and returns one whith those values replaced with valid values.

  Attributes
  ----------
  attribute_dict: dict
    The keys are the attributes of the class while the values are the default values. It's done this way rather than defined in the __init__ because this dictionary also defines what values need to be saved when written to disk, and what values need to be displayed when printed to the terminal.
  required_params: set of strs
    The parameters that must be provided to the transform at definition.
  cols : list of strs
    The column names of the array or dataframe to be transformed.
  num_examples : int
    The number of rows that have passed through the calc_global_values function.
  is_calc_run : bool
    Whether or not calc_global_values has been run for this transform.
  mean : numpy array
    The stored mean values to be used to normalize data.
  std : numpy array
    The stored standard deviation values to be used to normalize data.
  min : numpy array
    The stored mins to be used to normalize data.
  max : numpy array
    The stored maxes to be used to normalize data.


  """

  attribute_dict = {'norm_mode': None, 'norm_axis': None, 'fill_nan_func': lambda array: np.array(0), 'name': '', 'mean': None, 'std': None, 'min': None, 'max': None}

  for k, v in n.Transform.attribute_dict.iteritems():
    if k in attribute_dict:
      continue
    attribute_dict[k] = v

  required_params = set([])
  required_params.update(n.Transform.required_params)

  def __init__(self, from_file=None, save_dict=None, **kwargs):
    """Define a transform using a dictionary, file, or by setting the attribute values in kwargs.

    Parameters
    ----------
    from_file : None or str
      The file path of the Tranform that was written to disk.
    save_dict : dict or None
      The dictionary of attributes that completely define a Transform.
    name : str
      The name of the transform.
    dtype : numpy dtype
      The data type the transformed data should have. Defaults to np.float64.
    input_dtype: numpy dtype
      The datatype of the original inputted array.
    norm_mode : "mean_std", "min_max" or None
      How to normalize the data. Subtracting out the mean and dividing by the standard deviation; Subtracting out min and dividing by the difference between max and min; or leaving as is.
    norm_axis : 0, 1, or None
      In the case that 'norm_mode' has a non None value, what axis should be used for the normalization. None implies both 0 and 1.
    fill_nan_func : np.array(dtype=np.float64) -> np.array(dtype=np.float64)
      Function which fills in NaN values with some value. This function takes in a numpy array which contains NaN and returns one whith those values replaced with valid values.
    **kwargs :
      The keyword arguments that set the values of the attributes defined in the attribute_dict.

    """
    super(NumTransform, self).__init__(from_file, save_dict, **kwargs)

    # Ensure a valid norm mode was passed
    valid_norm_modes = ('mean_std', 'min_max', None)
    if self.norm_mode not in valid_norm_modes:
      raise ValueError("{} is an invalid norm_mode. Accepted norm mods are ".format(self.norm_mode, valid_norm_modes))

    # Ensure a valid norm axis was passed
    valid_norm_axis = (0, 1, (0, 1), None)
    if self.norm_axis not in valid_norm_axis:
      raise ValueError("{} is an invalid norm_axis. Accepted norm axes are ".format(self.norm_axis, valid_norm_axis))

  def __len__(self):
    """Get the length of the transformed data"""
    # assert self.is_calc_run, ("Must run calc_global_values before taking the len.")
    return len(self.cols)

  def _calc_global_values(self, array):
    """Calculate all the values of the Transform that are dependent on all the examples of the dataset. (e.g. mean, standard deviation, unique category values, etc.) This method must be run before any actual transformation can be done.

    Parameters
    ----------
    array : np.ndarray
      Some of the data that will be transformed.

    """
    if self.input_dtype is None:
      self.input_dtype = array.input_dtype
    else:
      array = np.array(array, dtype=self.input_dtype)

    if self.dtype is None:
      self.dtype = self.input_dtype

    batch_size = float(array.shape[0])
    total_examples = self.num_examples + batch_size

    array = array.astype(self.dtype)

    if self.norm_mode == 'mean_std':
      # Find the means and standard deviations of each column
      array[np.isnan(array)] = self.fill_nan_func(array)
      self.mean = np.mean(array, axis=self.norm_axis)
      self.std = np.std(array, axis=self.norm_axis)

      if self.mean is None:
        self.mean = np.mean(array, axis=self.norm_axis)
        self.var = np.var(array, axis=self.norm_axis)
      else:
        self.mean = (self.num_examples / total_examples) * self.mean + (batch_size / total_examples) * np.mean(array, axis=self.norm_axis)
        self.var = np.var(array, axis=self.norm_axis)

    elif self.norm_mode == 'min_max':
      # Find the means and standard deviations of each column
      array[np.isnan(array)] = self.fill_nan_func(array)

      if self.min is None:
        self.min = np.min(array, axis=self.norm_axis)
        self.max = np.max(array, axis=self.norm_axis)
      else:
        self.min = np.minimum(self.min, np.min(array, axis=self.norm_axis))
        self.max = np.maximum(self.max, np.max(array, axis=self.norm_axis))

  def _finish_calc(self):
    """Finish up the calc global value process."""
    if self.norm_mode == 'mean_std':
      self.std = np.sqrt(self.var)
      # If there are any standard deviations of 0, replace them with 1's,
      # print out a warning.
      if len(self.std[self.std == 0]):
        zero_std_cat_vals = []
        for index in np.where(self.std == 0.0)[0]:
          zero_std_cat_vals.append(self.index_to_cat_val[index])

        logging.warn(self.name + " has zero-valued stds at " + str(zero_std_cat_vals) + " replacing with 1's")

        self.std[self.std == 0] = 1.0
    elif self.norm_mode == 'min_max':
      # Test to make sure that min and max are not equal. If they are replace
      # with default values.
      if (self.min == self.max).any():
        if self.max.shape:
          self.max[self.max == self.min] = self.max[self.max == self.min] + 1
        else:
          self.max = self.max + 1

        logging.warn("NumTransform " + self.name + " the same values for min and max, replacing with " + str(self.min) + " " + str(self.max) + " respectively.")

  def _get_array_attributes(self, prefix=''):
    """Get the dictionary that contain the original shapes of the arrays before being converted into tfrecord examples.

    Parameters
    ----------
    prefix : str
      Any additional prefix string/dictionary keys start with. Defaults to no additional prefix.

    Returns
    -------
    array_attributes : dict
      The dictionary with keys equal to those that are found in the Transform's example dicts and values are the shapes of the arrays of a single example.

    """
    att_dict = {}
    att_dict['nums'] = {
      'shape': list([len(self.cols)]),
      'tf_type': feat.select_tf_dtype(self.dtype),
      'size': feat.size_from_shape([len(self.cols)]),
      'feature_func': feat.select_feature_func(self.dtype),
      'np_type': self.dtype
    }
    att_dict['nans'] = {
      'shape': [len(self.cols)],
      'tf_type': tf.int64,
      'size': feat.size_from_shape([len(self.cols)]),
      'feature_func': feat._int_feat,
      'np_type': np.bool
    }

    att_dict = self._pre(att_dict, prefix)
    return att_dict

  def define_waterwork(self, array=empty, return_tubes=None, prefix=''):
    """Get the waterwork that completely describes the pour and pump transformations.

    Parameters
    ----------
    array : np.ndarray or empty
      The array to be transformed.

    Returns
    -------
    Waterwork
      The waterwork with all the tanks (operations) added, and names set.

    """
    # Replace all the NaN's with the inputted replace_with function.
    nans, nans_slots = td.isnan(array)
    nans_slots['a'].set_name('array')

    nums, _ = td.replace(
      nans['a'], nans['target'],
      slot_plugs={
        'replace_with': lambda z: self.fill_nan_func(z[self._pre('array', prefix)])
      },
      tube_plugs={
        'replace_with': np.array([]),
        'replaced_vals': np.array(np.nan)
      }
    )

    nums['replaced_vals'].set_name('replaced_vals')
    nums['mask'].set_name('nans')

    # Do any additional normalization
    if self.norm_mode == 'mean_std':
      nums, _ = td.sub(
        nums['target'], self.mean,
        tube_plugs={'a_is_smaller': False, 'smaller_size_array': self.mean}
      )
      nums, _ = td.div(
        nums['target'], self.std,
        tube_plugs={'a_is_smaller': False, 'smaller_size_array': self.std, 'missing_vals': np.array([]), 'remainder': np.array([])}
      )
    elif self.norm_mode == 'min_max':
      nums, _ = td.sub(
        nums['target'], self.min,
        tube_plugs={'a_is_smaller': False, 'smaller_size_array': self.min}
      )
      nums, _ = td.div(
        nums['target'], (self.max - self.min),
        tube_plugs={'a_is_smaller': False, 'smaller_size_array': (self.max - self.min), 'missing_vals': np.array([]), 'remainder': np.array([])}
      )

    nums['target'].set_name('nums')

    if return_tubes is not None:
      ww = nums['target'].waterwork
      r_tubes = []
      for r_tube_key in return_tubes:
        r_tubes.append(ww.maybe_get_tube(r_tube_key))
      return r_tubes



"""Reshape tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Tile(ta.Tank):
  """Tile the elements of an array into an array with a shape defined by reps.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'tile'
  slot_keys = ['a', 'reps']
  tube_keys = ['target', 'old_shape', 'reps']
  pass_through_keys = ['reps']

  def _pour(self, a, reps):
    """Execute the Tile tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The array to reshape
    reps : list of ints
      The number of times to tile the array in each dimension

    Returns
    -------
    dict(
      target: np.ndarray
        The reshaped array
      old_shape: list of ints
        The old shape of the array
      reps : list of ints
        The number of times to tile the array in each dimension
    )

    """
    a = np.array(a)
    old_shape = a.shape
    target = np.tile(a, reps)

    return {'target': target, 'old_shape': old_shape, 'reps': reps}

  def _pump(self, target, old_shape, reps):
    """Execute the Tile tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The reshaped array
    old_shape: list of ints
      The old shape of the array
    reps : list of ints
      The number of times to tile the array in each dimension

    Returns
    -------
    dict(
      a: np.ndarray
        The array to reshape
      reps : list of ints
        The number of times to tile the array in each dimension
    )

    """
    shape = target.shape
    slice_indices = [0] * (len(shape) - len(old_shape))

    for dim in old_shape:
      slice_indices.append(slice(0, dim))

    a = target[tuple(slice_indices)]

    return {'a': a, 'reps': reps}



"""Sub tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Sub(ta.Tank):
  """The defintion of the Sub tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'sub'
  slot_keys = ['a', 'b']
  tube_keys = ['target', 'smaller_size_array', 'a_is_smaller']

  def _pour(self, a, b):
    """Execute the Sub tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: np.ndarray
      The object to subtract something from.
    b: np.ndarray
      The object which substracts from something else.

    Returns
    -------
    dict(
      target: np.ndarray
        The result of a-b.
      smaller_size_array: np.ndarray
        Either 'a' or 'b' depending on which has fewer elements.
      a_is_smaller: bool
        Whether or not 'a' is the smaller size array.
    )

    """
    # Convert to nump arrays
    if type(a) is not np.ndarray:
      a = np.array(a)
    if type(b) is not np.ndarray:
      b = np.array(b)

    # Copy whichever has a fewer number of elements and pass as output
    a_is_smaller = a.size < b.size
    if a_is_smaller:
      smaller_size_array = ut.maybe_copy(a)
    else:
      smaller_size_array = ut.maybe_copy(b)

    target = np.array(a - b)

    return {'target': target, 'smaller_size_array': smaller_size_array, 'a_is_smaller': a_is_smaller}

  def _pump(self, target, smaller_size_array, a_is_smaller):
    """Execute the Sub tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray
      The result of a-b.
    smaller_size_array: np.ndarray
      Either 'a' or 'b' depending on which has fewer elements.
    a_is_smaller: bool
      Whether or not 'a' is the smaller size array.

    Returns
    -------
    dict(
      a: np.ndarray
        The object to subtract something from.
      b: np.ndarray
        The object which substracts from something else.
    )

    """
    # Reconstruct the larger array from the smaller size array nd the target.
    if a_is_smaller:
      a = ut.maybe_copy(smaller_size_array)
      b = np.array(a - target)
    else:
      a = np.array(target + smaller_size_array)
      b = ut.maybe_copy(smaller_size_array)

    return {'a': a, 'b': b}



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestTile(th.TestTank):

  def test_scalar(self):
    a = np.array(3)
    reps = (1, 2)
    target = np.array([[3, 3]])
    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )

  def test_one_d(self):
    a = np.array([1, 2, 3])
    reps = ()
    target = np.array([1, 2, 3])
    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )

    a = np.array([1, 2, 3])
    reps = (1)
    target = np.array([1, 2, 3])
    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )
    a = np.array([1, 2, 3])
    reps = (2, 1)
    target = np.array([[1, 2, 3], [1, 2, 3]])
    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )

    a = np.array([1, 2, 3])
    reps = (2, 2, 1)
    target = np.array([
      [[1, 2, 3], [1, 2, 3]],
      [[1, 2, 3], [1, 2, 3]]
    ])
    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )

  def test_two_d(self):
    a = np.array([[1, 2, 3], [4, 5, 6]])
    reps = (1, 2)
    target = np.array([[1, 2, 3, 1, 2, 3], [4, 5, 6, 4, 5, 6]])

    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )

    a = np.array([[1, 2, 3], [4, 5, 6]])
    reps = (2, 1, 3)
    target = np.array([[[1, 2, 3, 1, 2, 3, 1, 2, 3], [4, 5, 6, 4, 5, 6, 4, 5, 6]], [[1, 2, 3, 1, 2, 3, 1, 2, 3], [4, 5, 6, 4, 5, 6, 4, 5, 6]]])

    self.pour_pump(
      td.tile,
      {
        'a': a,
        'reps': reps
      },
      {
        'target': target,
        'old_shape': a.shape,
        'reps': reps
      },
      test_type=False
    )
if __name__ == "__main__":
    unittest.main()



"""IterList and IterDict tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class IterList(ta.Tank):
  """The defintion of the IterList tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = '_iter_list'
  slot_keys = ['a']
  tube_keys = None

  def _pour(self, a):
    """Execute the IterList tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: list
      The tube list whose elements will be converted to tubes.

    Returns
    -------
    dict(
      a: list of values
    )

    """
    r_dict = {}
    for key_num, tube_key in enumerate(self.tube_keys):
      r_dict[tube_key] = a[key_num]
    return r_dict

  def _pump(self, **kwargs):
    """Execute the IterList tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    kwargs: list of values

    Returns
    -------
    dict(
      a: list
        The tube list whose elements will be converted to tubes.
    )

    """
    a = []
    for tube_key in self.tube_keys:
      a.append(kwargs[tube_key])
    return {'a': a}

  def _save_dict(self):
    save_dict = {}
    save_dict['func_name'] = self.func_name
    save_dict['name'] = self.name
    # save_dict['slots'] = [s for s in self.get_slots()]
    # save_dict['tubes'] = [t for t in self.get_tubes()]
    save_dict['kwargs'] = {'num_entries': len(self.tube_keys)}
    return save_dict

class IterDict(ta.Tank):
  """The defintion of the IterDict tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = '_iter_dict'
  slot_keys = ['a']
  tube_keys = None

  def _pour(self, a):
    """Execute the IterDict tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a: dict
      The tube dictionary. whose values will be converted to tubes.

    Returns
    -------
    dict(
      a: dict
        The dictionary of values
    )

    """
    r_dict = {}
    for tube_key in self.tube_keys:
      r_dict[tube_key] = a[tube_key]
    return r_dict

  def _pump(self, **kwargs):
    """Execute the IterDict tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    kwargs: dict
      The dictionary of values

    Returns
    -------
    dict(
      a: dict
        The tube dictionary. whose values will be converted to tubes.
    )

    """
    return {'a': kwargs}

  def _save_dict(self):
    save_dict = {}
    save_dict['func_name'] = self.func_name
    save_dict['name'] = self.name
    # save_dict['slots'] = [s for s in self.get_slots()]
    # save_dict['tubes'] = [t for t in self.get_tubes()]
    save_dict['kwargs'] = {'keys': self.tube_keys}
    return save_dict



"""Add tank definition."""
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class Add(ta.Tank):
  """The defintion of the CatToIndex tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """
  func_name = 'add'
  slot_keys = ['a', 'b']
  tube_keys = ['target', 'smaller_size_array', 'a_is_smaller']

  def _pour(self, a, b):
    """Execute the CatToIndex tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    cats: np.ndarray
      The array with all the category values to map to indices.
    cat_to_index_map: dict
      The mapping from category value to index. Must be one to one and contain all indices from zero to len(cat_to_index_map) - 1

    Returns
    -------
    dict(
      target: np.ndarray of ints
        The indices of all the corresponding category values from 'cats'.
      cat_to_index_map: dict
        The mapping from category value to index. Must be one to one and contain all indices from zero to len(cat_to_index_map) - 1
      missing_vals: list of category values
        All the category values from 'cats' which were not found in cat_to_index_map.
      input_dtype: a numpy dtype
        The dtype of the inputted 'cats' array.
    )

    """

    # Convert to nump arrays
    if type(a) is not np.ndarray:
      a = np.array(a)
    if type(b) is not np.ndarray:
      b = np.array(b)

    # Copy whichever has a fewer number of elements and pass as output
    a_is_smaller = a.size < b.size
    if a_is_smaller:
      smaller_size_array = ut.maybe_copy(a)
    else:
      smaller_size_array = ut.maybe_copy(b)

    target = np.array(a + b)

    return {'target': target, 'smaller_size_array': smaller_size_array, 'a_is_smaller': a_is_smaller}

  def _pump(self, target, smaller_size_array, a_is_smaller):
    """Execute the CatToIndex tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: np.ndarray of ints
      The indices of all the corresponding category values from 'cats'.
    cat_to_index_map: dict
      The mapping from category value to index. Must be one to one and contain all indices from zero to len(cat_to_index_map) - 1
    missing_vals: list of category values
      All the category values from 'cats' which were not found in cat_to_index_map.
    input_dtype: a numpy dtype
      The dtype of the inputted 'cats' array.

    Returns
    -------
    dict(
      cats: np.ndarray
        The array with all the category values to map to indices.
      cat_to_index_map: dict
        The mapping from category value to index. Must be one to one and contain all indices from zero to len(cat_to_index_map) - 1
    )

    """
    # reconstruct the other array from the smaller size array nd the target.
    if a_is_smaller:
      a = ut.maybe_copy(smaller_size_array)
      b = np.array(target - a)
    else:
      a = np.array(target - smaller_size_array)
      b = ut.maybe_copy(smaller_size_array)

    return {'a': a, 'b': b}



import unittest
import wtrwrks.utils.test_helpers as th
import wtrwrks.tanks.tank_defs as td
import numpy as np


class TestDiv(th.TestTank):
  def test_scalar(self):
    self.pour_pump(
      td.mul,
      {'a': 1, 'b': 2},
      {
        'target': np.array(2),
        'smaller_size_array': 2,
        'a_is_smaller': False,
        'missing_vals': np.array([], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int},
      test_type=False
    )
    self.pour_pump(
      td.mul,
      {'a': 2, 'b': 1},
      {
        'target': np.array(2),
        'smaller_size_array': 1,
        'a_is_smaller': False,
        'missing_vals': np.array([], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int},
      test_type=False
    )

    self.pour_pump(
      td.mul,
      {'a': 0, 'b': 1},
      {
        'target': np.array(0),
        'smaller_size_array': 1,
        'a_is_smaller': False,
        'missing_vals': np.array([0], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int},
      test_type=False
    )
    self.pour_pump(
      td.mul,
      {'a': 1, 'b': 0},
      {
        'target': np.array(0),
        'smaller_size_array': 0,
        'a_is_smaller': False,
        'missing_vals': np.array([1], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int},
      test_type=False
    )

  def test_array_int(self):
    self.pour_pump(
      td.mul,
      {'a': np.array([1, 2]), 'b': 2},
      {
        'target': np.array([2, 4]),
        'smaller_size_array': 2,
        'a_is_smaller': False,
        'missing_vals': np.array([], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int}
    )
    self.pour_pump(
      td.mul,
      {'b': np.array([1, 2]), 'a': 2},
      {
        'target': np.array([2, 4]),
        'smaller_size_array': 2,
        'a_is_smaller': True,
        'missing_vals': np.array([], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int}
    )
    self.pour_pump(
      td.mul,
      {'b': np.array([1, 3]), 'a': 2},
      {
        'target': np.array([2, 6]),
        'smaller_size_array': 2,
        'a_is_smaller': True,
        'missing_vals': np.array([], dtype=int),
      },
      type_dict={'a': np.ndarray, 'b': int}
    )

  def test_int_array(self):
    self.pour_pump(
      td.mul,
      {'a': 2., 'b': np.array([1., 2.])},
      {
        'target': np.array([2., 4.]),
        'smaller_size_array': 2.,
        'a_is_smaller': True,
        'missing_vals': np.array([]),
      },
      type_dict={'a': int, 'b': np.ndarray}
    )

    self.pour_pump(
      td.mul,
      {'a': 2., 'b': np.array([1, 0])},
      {
        'target': np.array([2., 0.]),
        'smaller_size_array': 2.,
        'a_is_smaller': True,
        'missing_vals': np.array([0], dtype=int),
      },
      type_dict={'a': int, 'b': np.ndarray},
      test_type=False
    )

    self.pour_pump(
      td.mul,
      {'a': 0., 'b': np.array([1, 2])},
      {
        'target': np.array([0., 0.]),
        'smaller_size_array': 0.,
        'a_is_smaller': True,
        'missing_vals': np.array([1, 2]),
      },
      type_dict={'a': int, 'b': np.ndarray},
      test_type=False
    )

    self.pour_pump(
      td.mul,
      {'b': 0., 'a': np.array([[1, 2], [3, 4]])},
      {
        'target': np.array([[0., 0.], [0., 0.]]),
        'smaller_size_array': 0.,
        'a_is_smaller': False,
        'missing_vals': np.array([1, 2, 3, 4]),
      },
      type_dict={'a': int, 'b': np.ndarray},
      test_type=False
    )

    self.pour_pump(
      td.mul,
      {'b': 2., 'a': np.array([[1, 0], [-1, 2]])},
      {
        'target': np.array([[2, 0.0], [-2, 4.]]),
        'smaller_size_array': 2.,
        'a_is_smaller': False,
        'missing_vals': np.array([0.]),
      },
      type_dict={'a': int, 'b': np.ndarray},
      test_type=False
    )

    self.pour_pump(
      td.mul,
      {'b': 0., 'a': np.array([-1, 2])},
      {
        'target': np.array([0., 0.]),
        'smaller_size_array': 0.,
        'a_is_smaller': False,
        'missing_vals': np.array([-1, 2]),
      },
      type_dict={'a': int, 'b': np.ndarray},
      test_type=False
    )

  def test_array(self):
    self.pour_pump(
      td.mul,
      {'a': np.array([1., 2.]), 'b': np.array([3., 4.])},
      {
        'target': np.array([3., 8.]),
        'smaller_size_array': np.array([3., 4.]),
        'a_is_smaller': False,
        'missing_vals': np.array([]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    self.pour_pump(
      td.mul,
      {'a': np.array([1.]), 'b': np.array([3., 4.])},
      {
        'target': np.array([3., 4.]),
        'smaller_size_array': np.array([1.]),
        'a_is_smaller': True,
        'missing_vals': np.array([]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    self.pour_pump(
      td.mul,
      {'a': np.array([0., 2.]), 'b': np.array([3., 0.])},
      {
        'target': np.array([0., 0.]),
        'smaller_size_array': np.array([3., 0.]),
        'a_is_smaller': False,
        'missing_vals': np.array([0., 2.]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    self.pour_pump(
      td.mul,
      {'a': np.array([0., 0.]), 'b': np.array([-3., 0.])},
      {
        'target': np.array([0.0, 0.]),
        'smaller_size_array': np.array([-3., 0.]),
        'a_is_smaller': False,
        'missing_vals': np.array([0., 0.]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    self.pour_pump(
      td.mul,
      {'a': np.array([0.]), 'b': np.array([[3., 4.]])},
      {
        'target': np.array([[0.0, 0.0]]),
        'smaller_size_array': np.array([0.]),
        'a_is_smaller': True,
        'missing_vals': np.array([3., 4.]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    self.pour_pump(
      td.mul,
      {'a': np.array([1.]), 'b': np.array([0., 4.])},
      {
        'target': np.array([0., 4.]),
        'smaller_size_array': np.array([1.]),
        'a_is_smaller': True,
        'missing_vals': np.array([0.]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )

    self.pour_pump(
      td.mul,
      {'b': np.array([0.]), 'a': np.array([[3., 0.0]])},
      {
        'target': np.array([[0., 0.]]),
        'smaller_size_array': np.array([0.]),
        'a_is_smaller': False,
        'missing_vals': np.array([3.0, 0.0]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )
    self.pour_pump(
      td.mul,
      {'b': np.array([1.]), 'a': np.array([0., 4.])},
      {
        'target': np.array([0., 4.0]),
        'smaller_size_array': np.array([1.]),
        'a_is_smaller': False,
        'missing_vals': np.array([0.]),
      },
      type_dict={'a': np.ndarray, 'b': np.ndarray}
    )


if __name__ == "__main__":
    unittest.main()



"""Shape tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import numpy as np
import wtrwrks.tanks.utils as ut

class Shape(ta.Tank):
  """Get the shape of an array.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = 'shape'
  slot_keys = ['a']
  tube_keys = ['target', 'a']
  pass_through_keys = ['a']

  def _pour(self, a):
    """

    Parameters
    ----------
    a: np.ndarray
      The array to get the shape of

    Returns
    -------
    dict(
      target: list of ints
        The shape of the array.
      a: np.ndarray
        The array to get the shape of
    )

    """
    return {'target': list(a.shape), 'a': ut.maybe_copy(a)}

  def _pump(self, target, a):
    """Execute the Shape tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    target: list of ints
      The shape of the array.
    a: np.ndarray
      The array to get the shape of

    Returns
    -------
    dict(
      a: np.ndarray
        The array to get the shape of
    )

    """
    return {'a': ut.maybe_copy(a)}



"""IterList and IterDict tank definition."""
import wtrwrks.waterworks.waterwork_part as wp
import wtrwrks.waterworks.tank as ta
import wtrwrks.tanks.utils as ut
import numpy as np


class TubeList(ta.Tank):
  """The defintion of the IterList tank. Contains the implementations of the _pour and _pump methods, as well as which slots and tubes the waterwork objects will look for.

  Attributes
  ----------
  slot_keys: list of strs
    The names off all the tank's slots, i.e. inputs in the pour (forward) direction, or outputs in the pump (backward) direction
  tubes: list of strs
    The names off all the tank's tubes, i.e. outputs in the pour (forward) direction,

  """

  func_name = '_tube_list'
  slot_keys = None
  tube_keys = ['target']

  def _pour(self, **kwargs):
    """Execute the IterList tank (operation) in the pour (forward) direction.

    Parameters
    ----------
    a1: object
      first element of the list
    .
    .
    .
    an: object
      nth element of the list

    Returns
    -------
    dict(
      a: list
        list of a1...an
    )

    """
    r_list = []
    for key in self.slot_keys:
      r_list.append(kwargs[key])
    return {'target': r_list}

  def _pump(self, target):
    """Execute the IterList tank (operation) in the pump (backward) direction.

    Parameters
    ----------
    a: list
      a: list of a1...an

    Returns
    -------
    dict(
      a1: object
        first element of the list
      .
      .
      .
      an: object
        nth element of the list
    )

    """
    r_dict = {}

    for num, slot_key in enumerate(self.slot_keys):
      r_dict[slot_key] = target[num]

    return r_dict

  def _save_dict(self):
    save_dict = {}
    save_dict['func_name'] = self.func_name
    save_dict['name'] = self.name
    # save_dict['slots'] = [s for s in self.get_slots()]
    # save_dict['tubes'] = [t for t in self.get_tubes()]
    save_dict['kwargs'] = {s: None for s in self.slot_keys}
    return save_dict
